# daq_driver.py

import time

import nidaqmx
import nidaqmx.task
import nidaqmx.task.channels
from nidaqmx._base_interpreter import BaseInterpreter
from qcodes import Instrument
from qcodes import validators as vals
from qcodes.instrument.channel import InstrumentChannel


class Daq(Instrument):
    """QCoDeS instrument driver for the National Instruments DAQ.

    Attributes:
    ---------
    name:
        The user-defined name for the instrument.
    address:
        The equipment address of the device, most commonly 'Dev1'.
    device:
        Connects the provided equipment address to the proper device in the
        NIDAQ system.
    ao_num:
        The number of AO channels.
    ai_num:
        The number of AI channels.
    max_out:
        The maximum voltage the device can produce.
    min_out:
        The minimum voltage the device can produce.
    max_in:
        The maximum voltage the device can receive.
    min_in:
        The minimum voltage the device can receive.
    """

    def __init__(self, name="Daq", address=""):
        """Initialization for the DAQ driver.

        Takes in the machine given device address (typically "Dev1"), and a user-defined
        name. Creates QCoDeS Parameters for each of the I/O channels.
        """
        start_time = time.time()
        # Initialize the DAQ system
        super().__init__(name)
        self.system = nidaqmx.system.System.local()
        self._address = address
        try:
            self.device = self.system.devices[self._address]
            # Grab the number of AO, AI channels
            self.ao_num = len(self.device.ao_physical_chans.channel_names)
            self.ai_num = len(self.device.ai_physical_chans.channel_names)
        except nidaqmx.errors.DaqError as e:
            self.close()
            raise e
        self._interpreter = BaseInterpreter
        # Read the max/min output/input voltages
        self.max_out = max(self.device.ao_voltage_rngs)
        self.min_out = min(self.device.ao_voltage_rngs)
        self.max_in = max(self.device.ai_voltage_rngs)
        self.min_in = min(self.device.ai_voltage_rngs)

        # For each output channel, create a corresponding DaqAOChannel object
        # If the device is already in use, system will throw a 'DaqError'. If so,
        # let's catch it and close the device, removing the instance from qcodes,
        # then throw the error.
        try:
            for a in range(self.ao_num):
                ch_name = "ao" + str(a)
                channel = DaqAOChannel(
                    self,
                    self.device,
                    self._address,
                    ch_name,
                    self.min_out,
                    self.max_out,
                    self.min_in,
                    self.max_in,
                )
                self.add_submodule(ch_name, channel)
            # Similarly, create a DaqAIChannel object for each (real) input channel
            count = 0
            for b in range(self.ai_num):
                # NI DAQs can use each of the coax pins as analog inputs separately, and
                # numbers them as such. (AI0-7 is voltage-part of coax reading for first
                # 8 pins, 8-15 is ground reading of coax for first 8 pins).
                # We don't want this behavior, we use each port as one input (ground and
                # voltage), so we make sure to only grab the "real" analog inputs

                if int(b / 8) % 2 == 0:
                    ch_name = "ai" + str(b)
                    channel = DaqAIChannel(
                        self,
                        self.device,
                        self._address,
                        ch_name,
                        self.min_in,
                        self.max_in,
                    )
                    self.add_submodule(ch_name, channel)
                    count += 1
            # Update the actual number of ai ports
            self.ai_num = count
        except nidaqmx.errors.DaqError as e:
            self.close()
            raise e

        self.connect_message()

    def get_idn(self):
        vendor = "National Instruments"
        model = f"DAQ {self.device.product_type}"
        serial = self.device.dev_serial_num
        firmware = ".".join([str(x) for x in self.system.driver_version])

        return dict(
            zip(
                ("vendor", "model", "serial", "firmware"),
                [vendor, model, serial, firmware],
            )
        )

    def get_ao_num(self):
        """Returns the number of AO ports available."""
        return self.ao_num

    def get_ai_num(self):
        """Returns the number of AI ports available."""
        return self.ai_num

    def update_all_inputs(self):
        """Updates all the AI channel voltage values."""
        for chan_name in self.submodules:
            self.submodules[chan_name].get("voltage")

    def snapshot_base(self, update=False, params_to_skip_update=None):
        """QCoDeS method to obstain the state of the instrument as a
        JSON-compatible dictionary.

        Supported by the custom JSON encoder class:
        qcodes.utils.helpers.NumpyJSONEncoder

        Parameters
        ---------
        update:
            If True, update the state by querying the instrument.
            If None, only update if the state is known to be invalid.
            If False, use the latest values in memory and do not update.
        params_to_skip_update:
            List of parameter names that will be skipped in update even if
            update is True. This is useful if you have parameters that are
            slow to update but can be updated in a different way (as in the qdac).
            If you want to skip the update of certain parameters in all snapshots,
            use the `snapshot_get` attribute of those parameters instead.

        Returns:
        ---------
            Dictionary containing snapshot of information on base instrument
        """
        snap = super().snapshot_base(
            update=update, params_to_skip_update=params_to_skip_update
        )
        snap["address"] = self._address

        return snap

    def close(self):
        """Stops the QCoDeS instrument frees its resources.

        Closes all subclasses which may have other resources to close;
        writes the object configuration in the desired filepath.
        """
        if hasattr(self, "submodules"):
            for a, c in self.submodules.items():
                c.close()

        super().close()

    def __del__(self):
        """Destructor method to close program.

        Necessary for the nidaqmx library to avoid issues upon
        relaunching the software.
        """
        try:
            self.close()
        except Exception as e:
            print("Driver couldn't close properly.")
            print(e)

        try:
            for a, c in self.submodules.items():
                # Removes all Task objects from the channels, so system doesn't complain
                c.__del__()
        except:
            pass

        # Makes sure the Instrument destructor is called
        super().__del__()


class DaqAOChannel(InstrumentChannel):
    """Child class to represent AO channels for DAQ.

    Defines utility functions for each Parameter. Channels are created
    when DAQ Instrument is initialized.

    Attributes:
    ---------
    parent:
        QCoDeS parent instrument, in this case the DAQ.
    device:
        Device address of the machine.
    address:
        Unique address used to connect to DAQ instrument.
    channel:
        Channel name for individual AO channels.
    my_name:
        Stores individual channel names.
    fullname:
        Device address combined with channel name.
    """

    def __init__(
        self,
        parent: Instrument,
        device,
        address,
        channel,
        min_output,
        max_output,
        min_in,
        max_in,
    ):
        """Initialization for the DAQ output channels.

        Saves information about the DAQ, creates and initializes
        the QCoDeS values.
        """
        # Saves the information about the DAQ this channel belongs to
        super().__init__(parent, channel)
        self.device = device
        self.address = address
        self.channel = channel
        # Name, e.g. ao1
        self.my_name = str(channel)
        # Full name, e.g. Dev1/ao1
        self.fullname = self.address + "/" + self.my_name

        # Extrema
        self.max_V = max_output
        self.min_V = min_output
        self.min_in = min_in
        self.max_in = max_in

        self.write_task = None
        self.read_task = None
        self.channel = None
        self.channel_handle = None
        self.create_tasks()

        # Initializes the values of the Parameters
        self.gain = 1
        self._voltage = self.get_voltage()
        self.impedance = None
        self._value = 0

        # Add a Parameter for the output gain, set the unit to 'V/V' as default
        # self.add_parameter('gain_factor',
        #                    get_cmd=self.get_gain_factor,
        #                    set_cmd=self.set_gain_factor,
        #                    label=f'{self.my_name} Gain factor',
        #                    unit='V/V'
        #                    )

        # Create the Parameter for the units of gain
        # self.add_parameter('gain_units',
        #                   set_cmd=self.set_gain_units,
        #                   label=f'{self.my_name} Gain units',
        #                   )

        # Add a Parameter for the output impedance
        # self.add_parameter('impedance',
        #                   get_cmd=self.get_load_impedance,
        #                   get_parser=float,
        #                   set_cmd=self.set_load_impedance,
        #                   label=f'{self.my_name} Load Impedance',
        #                   vals=vals.Numbers(0, 1000)
        #                   )

        # Add a Parameter for the value
        # self.add_parameter('value',
        #                   get_cmd=self.get_value,
        #                   set_cmd=self.set_value,
        #                   label='Voltage * Factor'
        #                   )

        # Add a Parameter for the voltage
        self.add_parameter(
            "voltage",
            get_cmd=self.get_voltage,
            get_parser=float,
            set_cmd=self.set_voltage,
            label=f"{self.my_name} Voltage",
            unit="V",
            vals=vals.Numbers(self.min_V, self.max_V),
        )

    def get_gain_factor(self):
        """Returns the gain value and the unit conversion (e.g., mA/V) as a tuple."""
        return (self.gain, self.parameters["gain factor"].unit)

    def set_gain_factor(self, _gain):
        """Sets the gain value.

        Parameters
        ---------
        _gain:
            Value to be set as gain attribute.
        """
        self.gain = _gain

    #        if self.channel != None:
    #            self.channel.ai_gain=_gain

    def set_gain_units(self, units):
        """Sets the gain unit conversion (e.g., uV/V).

        Parameters
        ---------
        units:
            Value of desired final units of gain.
            Default unit is the Volt.
        """
        self.parameters["gain factor"].unit = units

    def get_voltage(self):
        """Returns the current output voltage."""
        if self.read_task is not None:
            self._voltage = self.read_task.read()

        return self._voltage

    def set_voltage(self, _voltage):
        """Sets the voltage to the desired value."""
        if self.write_task is not None:
            self.write_task.write(_voltage)

            self._voltage = _voltage

    def get_value(self):
        """Calculates, sets, and returns the value after the gain.

        Returns:
        ---------
            Product of gain and voltage with associated units.
        """
        parts = self.parameters["gain factor"].unit.split("/")
        self.parameters["value"].unit = parts[0]
        self._value = self.gain * self.get_voltage()
        return (self._value, self.parameters["value"].unit)

    def set_value(self, value):
        """Sets the voltage to the appropriate value after considering the gain.

        Parameters
        ---------
        value:
            The product of voltage and gain.
        """
        self._value = value
        self.set_voltage(self._value / self.gain)

    def get_load_impedance(self):
        """Returns the load impedance."""
        return self.impedance

    def set_load_impedance(self, _imp):
        """Sets the load impedance."""
        self.impedance = _imp

    #        if self.channel != None:
    #            self.channel.ao_load_impedance=_imp

    def create_tasks(self):
        """Defines the tasks that should be used to conduct reads and writes to the port."""
        if self.write_task is not None or self.read_task is not None:
            self.clear_task()

        # We now create a task to read and write on each channel
        self.write_task = nidaqmx.Task(f"writing {self.address}_{self.my_name}")
        self.read_task = nidaqmx.Task(f"reading {self.address}_{self.my_name}")

        # Add the channel to the task
        self.write_task.ao_channels.add_ao_voltage_chan(self.fullname)
        self.read_task.ai_channels.add_ai_voltage_chan(
            f"{self.address}/_{self.my_name}_vs_aognd",
            min_val=self.min_in,
            max_val=self.max_in,
        )

        # Channel handler that can be used to communicate things like gain, impedance
        # back to the DAQ
        self.channel_handle = nidaqmx.task.channels._ao_channel.AOChannel(
            self.write_task._handle, self.channel, BaseInterpreter
        )

    #        if self.gain != -1:
    #            task.ao_channels.ao_gain=self.gain
    #        if self.impedance != -1:
    #            task.ao_load_impedance=self.impedance

    def clear_task(self):
        """Clears the current task object for resource maintenance."""
        if self.read_task is not None:
            self.read_task.close()
        if self.write_task is not None:
            self.write_task.close()
        self.read_task = None
        self.write_task = None
        self.channel_handle = None

    def close(self):
        """Destructor, makes sure task is clean."""
        self.clear_task()


class DaqAIChannel(InstrumentChannel):
    """Child class to represent AO channels for DAQ.

    Defines utility functions for each Parameter. Channels are created
    when DAQ Instrument is initialized.

    Attributes:
    ---------
    parent:
        QCoDeS parent instrument, in this case the DAQ.
    device:
        Device address of the machine.
    address:
        Unique address used to connect to DAQ instrument.
    channel:
        Channel name for individual AI channels.
    my_name:
        Stores individual channel names.
    fullname:
        Device address combined with channel name.
    """

    def __init__(
        self, parent: Instrument, device, address, channel, min_input, max_input
    ):
        """Initialization for DAQ input channels.

        Saves information about the DAQ, creates and initializes
        the QCoDeS values.
        """
        # Saves the information about the DAQ this channel belongs to
        super().__init__(parent, channel)
        self.device = device
        self.address = address
        self.channel = channel
        self.my_name = str(channel)
        self.fullname = self.address + "/" + self.my_name

        # Set our maximum input/output
        self.max_in = max_input
        self.min_in = min_input

        self.task = None
        self.channel = None
        self.channel_handle = None
        self.create_task()

        # Set the default Parameter values
        self.gain = 1
        self._voltage = self.get_voltage()
        self.impedance = None
        self._value = 0

        # Create the Parameter for gain, with default unit 'V/V'
        # self.add_parameter('gain_factor',
        #                    get_cmd=self.get_gain_factor,
        #                    set_cmd=self.set_gain_factor,
        #                    label=f'{self.my_name} gain',
        #                    unit='V/V'
        #                    )

        # Create the Parameter for the units of gain
        # self.add_parameter('gain_units',
        #                   set_cmd=self.set_gain_units,
        #                   label=f'{self.my_name} gain units',
        #                   )

        # Create the Parameter for impedance
        # self.add_parameter('impedance',
        #                   get_cmd=self.get_load_impedance,
        #                   get_parser=float,
        #                   set_cmd=self.set_load_impedance,
        #                   label=f'{self.my_name} Load Impedance',
        #                   vals=vals.Numbers(0, 1000)
        #                   )

        # Create the Parameter for value, which is the voltage * gain
        # self.add_parameter('value',
        #                   get_cmd=self.get_value,
        #                   label='Voltage * Factor')

        # Create the Parameter for input voltage
        self.add_parameter(
            "voltage",
            get_cmd=self.get_voltage,
            get_parser=float,
            label=f"{self.my_name} Voltage",
            unit="V",
            vals=vals.Numbers(self.min_in, self.max_in),
        )

    def get_gain_factor(self):
        """Returns the gain factor unit conversion."""
        return (self.gain, self.parameters["gain factor"].unit)

    def set_gain_factor(self, _gain):
        """Sets the gain factor value.

        Parameters
        ---------
            _gain:
                Value to be set as gain attribute.
        """
        self.gain = _gain

    #        if self.channel != None:
    #            self.channel.ai_gain=_gain

    def set_gain_units(self, units):
        """Sets the gain unit conversion (e.g., uV/V).

        Parameters
        ---------
        units:
            Value of desired final units of gain.
            Default unit is the Volt.
        """
        self.parameters["gain"].unit = units

    def get_voltage(self):
        """Returns the current input voltage."""
        if self.task is not None:
            self._voltage = self.task.read()
        return self._voltage

    def get_value(self):
        """Calculates, sets, and returns the value after the gain.

        Returns:
        ---------
            Product of gain and voltage with associated units.
        """
        parts = self.parameters["gain factor"].unit.split("/")
        self.parameters["value"].unit = parts[0]
        self._value = self.gain * self.get_voltage()
        return (self._value, self.parameters["value"].unit)

    def get_load_impedance(self):
        """Returns the load impedance."""
        return self.impedance

    def set_load_impedance(self, _imp):
        """Sets the load impedance."""
        self.impedance = _imp

    #        if self.channel != None:
    #            self.channel.ai_load_impedance=_imp

    def create_task(self):
        """Defines the task that should be used to conduct reads from the port."""
        if self.task is not None:
            self.clear_task()

        self.task = nidaqmx.Task(f"reading {self.address}_{self.my_name}")

        # Add the channel to the task
        self.task.ai_channels.add_ai_voltage_chan(
            self.fullname, min_val=self.min_in, max_val=self.max_in
        )

        # Channel handler that can be used to communicate things like gain, impedance
        # back to the DAQ
        self.channel_handle = nidaqmx.task.channels._ai_channel.AIChannel(
            self.task._handle, self.channel, BaseInterpreter
        )
        #        if self.gain != -1:
        #            task.ai_channels.ai_gain=self.gain
        #        if self.impedance != -1:
        #            task.ai_load_impedance=self.impedance

    def clear_task(self):
        """Utility function to clear the task object for resource maintenance."""
        if self.task is not None:
            self.task.close()
        self.task = None
        self.channel_handle = None

    def close(self):
        """Destructor, makes sure task is clean."""
        self.clear_task()



from qcodes import VisaInstrument
from qcodes.utils.validators import Enum, Strings


class Keithley_2450(VisaInstrument):
    """QCoDeS driver for the Keithley 2450 voltage source.
    Jan-10-2019 2400 driver updated with parser corrected
    """

    def __init__(self, name, address, **kwargs):
        super().__init__(name, address, terminator="\n", **kwargs)

        self.add_parameter(
            "rangev",
            get_cmd="SENS:VOLT:RANG?",
            get_parser=float,
            set_cmd="SOUR:VOLT:RANG {:f}",
            label="Voltage range",
        )

        self.add_parameter(
            "rangei",
            get_cmd="SENS:CURR:RANG?",
            get_parser=float,
            set_cmd="SOUR:CURR:RANG {:f}",
            label="Current range",
        )

        self.add_parameter(
            "compliancev",
            get_cmd="SENS:VOLT:PROT?",
            get_parser=float,
            set_cmd="SENS:VOLT:PROT {:f}",
            label="Voltage Compliance",
        )

        self.add_parameter(
            "compliancei",
            get_cmd="SENS:CURR:PROT?",
            get_parser=float,
            set_cmd="SENS:CURR:PROT {:f}",
            label="Current Compliance",
        )

        self.add_parameter(
            "volt",
            get_cmd=":READ?",
            get_parser=self._volt_parser,
            set_cmd=":SOUR:VOLT:LEV {:.8f}",
            label="Voltage",
            unit="V",
        )

        self.add_parameter(
            "curr",
            get_cmd=":READ?",
            get_parser=self._curr_parser,
            set_cmd=":SOUR:CURR:LEV {:.8f}",
            label="Current",
            unit="A",
        )

        self.add_parameter(
            "mode",
            vals=Enum("VOLT", "CURR"),
            get_cmd=":SOUR:FUNC?",
            set_cmd=self._set_mode_and_sense,
            label="Mode",
        )

        self.add_parameter(
            "sense",
            vals=Strings(),
            get_cmd=":SENS:FUNC?",
            set_cmd=':SENS:FUNC "{:s}"',
            label="Sense mode",
        )

        self.add_parameter(
            "output", get_parser=int, set_cmd=":OUTP:STAT {:d}", get_cmd=":OUTP:STAT?"
        )

        self.add_parameter(
            "nplcv",
            get_cmd="SENS:VOLT:NPLC?",
            get_parser=float,
            set_cmd="SENS:VOLT:NPLC {:f}",
            label="Voltage integration time",
        )

        self.add_parameter(
            "nplci",
            get_cmd="SENS:CURR:NPLC?",
            get_parser=float,
            set_cmd="SENS:CURR:NPLC {:f}",
            label="Current integration time",
        )

        self.add_parameter(
            "resistance",
            get_cmd=":READ?",
            get_parser=self._resistance_parser,
            label="Resistance",
            unit="Ohm",
        )

    def _set_mode_and_sense(self, msg):
        # This helps set the correct read out curr/volt
        if msg == "VOLT":
            self.sense("CURR")
        elif msg == "CURR":
            self.sense("VOLT")
        else:
            raise AttributeError("Mode does not exist")
        self.write(f":SOUR:FUNC {msg:s}")

    def reset(self):
        """Reset the instrument. When the instrument is reset, it performs the
        following actions.

            Returns the SourceMeter to the GPIB default conditions.

            Cancels all pending commands.

            Cancels all previously send `*OPC` and `*OPC?`
        """
        self.write(":*RST")

    def _volt_parser(self, msg):
        fields = [float(x) for x in msg.split(",")]
        return fields[0]

    def _curr_parser(self, msg):
        fields = [float(x) for x in msg.split(",")]
        print(msg)
        return fields[0]

    def _resistance_parser(self, msg):
        fields = [float(x) for x in msg.split(",")]
        print(msg)
        return fields[0]



import os
from pathlib import Path
from typing import Dict

import pytest

import measureit
import measureit.config as config
import measureit.tools.util as util


@pytest.fixture(autouse=True)
def reset_measureit_config(monkeypatch):
    """Ensure each test starts with a clean configuration state."""
    config._DATA_DIR_OVERRIDE = None  # type: ignore[attr-defined]
    monkeypatch.delenv("MEASUREIT_HOME", raising=False)
    monkeypatch.delenv("MeasureItHome", raising=False)
    yield
    config._DATA_DIR_OVERRIDE = None  # type: ignore[attr-defined]


def _fake_user_data_dir(target: Path):
    return lambda *_, **__: str(target)


def test_get_path_uses_platformdirs_and_creates(tmp_path, monkeypatch):
    base = tmp_path / "platform_home"
    monkeypatch.setattr(config, "user_data_dir", _fake_user_data_dir(base))

    databases = config.get_path("databases")
    logs = config.get_path("logs")

    assert databases == base / "Databases"
    assert logs == base / "logs"
    assert databases.exists()
    assert logs.exists()


def test_get_path_respects_measurithome_env(tmp_path, monkeypatch):
    home = tmp_path / "env_home"
    monkeypatch.setenv("MEASUREIT_HOME", str(home))

    cfg_dir = config.get_path("cfg")
    assert cfg_dir == home / "cfg"
    assert cfg_dir.exists()


def test_get_path_respects_legacy_measureithome(tmp_path, monkeypatch):
    legacy_home = tmp_path / "legacy_home"
    monkeypatch.setenv("MeasureItHome", str(legacy_home))

    origin = config.get_path("origin_files")
    assert origin == legacy_home / "Origin Files"
    assert origin.exists()


def test_set_data_dir_overrides_environment(tmp_path, monkeypatch):
    env_home = tmp_path / "env_home"
    custom_home = tmp_path / "custom_home"
    monkeypatch.setenv("MEASUREIT_HOME", str(env_home))

    result = measureit.set_data_dir(custom_home)
    assert result == custom_home
    cfg_dir = measureit.get_path("cfg")
    assert cfg_dir == custom_home / "cfg"
    assert cfg_dir.exists()
    assert Path(measureit.get_path("databases")).parent == custom_home
    assert Path(measureit.get_path("logs")).parent == custom_home


def test_set_data_dir_updates_environment_variables(tmp_path):
    custom_home = tmp_path / "custom_env"
    measureit.set_data_dir(custom_home)

    assert measureit.get_path("databases").parent == custom_home
    assert measureit.get_path("logs").parent == custom_home
    assert os.environ["MEASUREIT_HOME"] == str(custom_home)
    assert os.environ["MeasureItHome"] == str(custom_home)


def test_get_path_rejects_unknown_subdir(monkeypatch):
    monkeypatch.setattr(
        config, "user_data_dir", _fake_user_data_dir(Path("/tmp/unused"))
    )
    with pytest.raises(ValueError):
        config.get_path("unknown")  # type: ignore[arg-type]


def test_database_creation_uses_configured_path(tmp_path, monkeypatch):
    base = tmp_path / "db_home"
    monkeypatch.setattr(config, "user_data_dir", _fake_user_data_dir(base))

    created: Dict[str, str] = {}

    def fake_init(path: str):
        created["path"] = path

    def fake_new_experiment(*_, **__):
        return None

    monkeypatch.setattr(util, "initialise_or_create_database_at", fake_init)
    monkeypatch.setattr(util.qc, "new_experiment", fake_new_experiment)

    util.init_database("example", "exp", "sample")

    expected_path = base / "Databases" / "example.db"
    assert created["path"] == str(expected_path)
    assert expected_path.parent.exists()
    expected_path.touch()
    assert expected_path.exists()



# sweep0d.py

import time

from PyQt5.QtCore import QObject

from ..tools.util import _autorange_srs, safe_get
from .base_sweep import BaseSweep
from .progress import SweepState


class Sweep0D(BaseSweep, QObject):
    """Class for the following/live plotting of one parameter against monotonic time.

    As of now, is just an extension of BaseSweep, but has been separated for
    future convenience. The default independent variable for Sweep0D Measurements
    is time. The measured data is transferred in real-time (through QObject slot
    and signal connections) from the Sweep classes to the Runner and Plotter Threads
    to organize, save, and live-plot the tracked parameters.

    Attributes:
    ---------
    max_time:
        The cutoff time (in seconds) where the sweep will end itself.
    *args:
        Used to set Runner and Plotter threads if previously created by GUI.
    **kwargs:
        Passes any keyword arguments to BaseSweep class.
    direction:
        Not utilized in Sweep0D.

    Methods:
    ---------
    flip_direction()
        Informs user that direction can not be flipped in Sweep0D.
    update_values()
        Returns dictionary of updated [parameter:value] pairs.
    """

    def __init__(self, max_time=1e6, *args, **kwargs):
        """Calls the BaseSweep initialization with any extra variables.

        Parameters
        ---------
        max_time:
            The cutoff time (in seconds) where the sweep will end itself.
        *args:
            Used to set Runner and Plotter threads if previously created by GUI.
        **kwargs:
            Passes any keyword arguments to BaseSweep class.
        direction:
            Not utilized in Sweep0D.
        """
        QObject.__init__(self)
        BaseSweep.__init__(self, set_param=None, *args, **kwargs)

        # Amount of time to run
        self.max_time = max_time
        self.progressState.progress = 0.0
        self.update_progress()

    def __str__(self):
        if self.max_time is None:
            return "Continuous 0D Sweep"
        else:
            return f"0D Sweep for {self.max_time} seconds."

    def __repr__(self):
        return f"Sweep0D({self.max_time}, {1.0 / self.inter_delay})"

    def flip_direction(self):
        """The independent variable can not be flipped in Sweep0D.

        The default independent variable is time, and can not be flipped.
        The function is defined so that it will not cause an error if called.
        """
        self.print_main.emit(
            "Can't flip the direction, as we are not sweeping a parameter."
        )
        return

    def update_values(self):
        """Obtains all current parameter values.

        Data is collected in parameter-value pairs, saved to a database if
        desired, and a signal is emitted sending this data to the slots of
        the sweep.

        Returns:
        ---------
        data:
            A list of tuples with the new data. Each tuple is of the format
            (<QCoDeS Parameter>, measurement value). The tuples are passed in order
            of time, set_param (if applicable), then all the followed parameters.
        """
        t = self.progressState.time_elapsed

        data = []

        if t >= self.max_time:
            if (
                self.save_data
                and self.runner is not None
                and self.runner.datasaver is not None
            ):
                self.runner.datasaver.flush_data_to_database()
            self.progressState.state = SweepState.DONE
            self.print_main.emit(f"Done with the sweep, t={t} (s)")
            self.completed.emit()

            return None
        else:
            data.append(("time", t))

        persist_param = None
        if self.persist_data is not None:
            data.append(self.persist_data)
            persist_param = self.persist_data[0]

        for i, (l, _, gain) in enumerate(self._srs):
            _autorange_srs(l, 3)

        for i, p in enumerate(self._params):
            if p is not persist_param:
                v = safe_get(p)
                data.append((p, v))

        if self.save_data and self.progressState.state == SweepState.RUNNING:
            self.runner.datasaver.add_result(*data)

        self.send_updates()

        return data

    def estimate_time(self, verbose=True):
        """Returns an estimate of the amount of time the sweep will take to complete.

        Parameters
        ----------
        verbose:
            Controls whether the function will print out the estimate in the form hh:mm:ss (default True)

        Returns:
        -------
        Time estimate for the sweep, in seconds
        """
        if self.progressState.state in (SweepState.READY, SweepState.RAMPING):
            remaining = self.max_time
        elif self.progressState.state == SweepState.DONE:
            remaining = 0
        else:
            remaining = max(self.max_time - self.progressState.time_elapsed, 0.0)

        if verbose:
            hours, minutes, seconds = self._split_hms(remaining)
            self.print_main.emit(
                f"Estimated time remaining for {repr(self)}: {hours}h:{minutes:02d}m:{seconds:02d}s"
            )

        return remaining

    # --- JSON export/import hooks ---
    def _export_json_specific(self, json_dict: dict) -> dict:
        json_dict["set_param"] = None
        json_dict["attributes"]["max_time"] = self.max_time
        return json_dict

    @classmethod
    def from_json(cls, json_dict, station):
        attrs = json_dict.get("attributes", {})
        # Sweep0D takes max_time plus common BaseSweep kwargs
        max_time = attrs.pop("max_time", 1e6)
        return cls(max_time=max_time, **attrs)



import pytest


def test_import_measureit_public_modules():
    # Canonical subpackage imports (no deprecation warnings)
    from measureit.sweep import (
        GateLeakage,
        SimulSweep,
        Sweep0D,
        Sweep1D,
        Sweep2D,
        SweepIPS,
    )

    assert all([Sweep0D, Sweep1D, Sweep2D, SimulSweep, SweepIPS, GateLeakage])
    # Tools
    import measureit.tools.tracking as _trk  # noqa: F401
    from measureit.tools import init_database  # noqa: F401
    from measureit.tools.sweep_queue import DatabaseEntry, SweepQueue  # noqa: F401


def test_import_qcodes_drivers_available():
    # Optional external drivers: skip test if not installed in environment
    pytest.importorskip("qcodes")

    m = pytest.importorskip("qcodes.instrument_drivers.stanford_research.SR830")
    assert hasattr(m, "SR830")
    m = pytest.importorskip("qcodes.instrument_drivers.stanford_research.SR860")
    assert hasattr(m, "SR860")
    m = pytest.importorskip("qcodes.instrument_drivers.tektronix.Keithley_2450")
    assert hasattr(m, "Keithley2450")
    m = pytest.importorskip("qcodes.instrument_drivers.tektronix.Keithley_2400")
    assert hasattr(m, "Keithley2400")


def test_optional_nidaqmx_import():
    # This package may not be installed; ensure import does not break the suite
    try:
        import nidaqmx  # noqa: F401
    except Exception:
        pytest.skip("nidaqmx not available in this environment")


def test_new_subpackage_paths_work():
    # New canonical imports via subpackages
    from measureit.sweep import (
        GateLeakage,
        SimulSweep,
        Sweep0D,
        Sweep1D,
        Sweep2D,
        SweepIPS,
    )

    assert Sweep0D and Sweep1D and Sweep2D and SimulSweep and SweepIPS and GateLeakage

    # Import tools subpackage minimal exports
    from measureit.tools import init_database, tracking  # noqa: F401

    assert init_database is not None
    # Import SweepQueue from its module to avoid circular import at package init
    from measureit.tools.sweep_queue import DatabaseEntry, SweepQueue  # noqa: F401

    assert SweepQueue and DatabaseEntry

    from measureit.visualization import (  # noqa: F401
        Heatmap,
        print_all_metadata,
        print_metadata,
    )

    assert Heatmap and print_metadata and print_all_metadata


def test_no_deprecations_with_canonical_imports():
    import warnings

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("error", category=DeprecationWarning)
        # Only subpackage imports should not warn
        from measureit.sweep import (
            GateLeakage,
            SimulSweep,
            Sweep0D,
            Sweep1D,
            Sweep2D,
            SweepIPS,
        )
        from measureit.tools import init_database, tracking  # noqa: F401
        from measureit.tools.sweep_queue import DatabaseEntry, SweepQueue
        from measureit.visualization import Heatmap  # noqa: F401

        assert all(
            [
                Sweep0D,
                Sweep1D,
                Sweep2D,
                SimulSweep,
                SweepIPS,
                GateLeakage,
                SweepQueue,
                DatabaseEntry,
                init_database,
            ]
        )



from qcodes import ChannelList, InstrumentChannel, VisaInstrument
from qcodes.instrument.group_parameter import Group, GroupParameter
from qcodes.utils.validators import Enum


class SensorChannel(InstrumentChannel):
    """Channel class for the Lakeshore 331.

    Args:
        parent: The parent Lakeshore 331.
        name: The channel name.
        channel: The channel ID.

    Attributes:
        channel: The channel ID.
    """

    _CHANNEL_VAL = Enum("A", "B")

    def __init__(self, parent, name, channel):
        super().__init__(parent, name)

        self._CHANNEL_VAL.validate(channel)
        self._channel = channel

        self.add_parameter(
            "temperature",
            get_cmd=f"KRDG? {self._channel}",
            get_parser=float,
            label="Temerature",
            unit="K",
        )

        self.add_parameter(
            "sensor_raw",
            get_cmd=f"SRDG? {self.channel}",
            get_parser=float,
            label=f"sensor raw {self.channel}",
        )

        self.add_parameter(
            "sensor_status",
            get_cmd=f"RDGST? {self._channel}",
            val_mapping={
                "OK": 0,
                "Invalid Reading": 1,
                "Temp Underrange": 16,
                "Temp Overrange": 32,
                "Sensor Units Zero": 64,
                "Sensor Units Overrange": 128,
            },
            label="Sensor_Status",
        )


class Lakeshore_331(VisaInstrument):
    """Instrument class for the Lakeshore 331.

    Args:
        name: The channel name.
        address: The GPIB address.
    """

    _loop = 1

    def __init__(self, name, address, **kwargs):
        super().__init__(name, address, terminator="\r\n", **kwargs)
        channels = ChannelList(self, "TempSensors", SensorChannel, snapshotable=False)
        for chan_name in ("A", "B"):
            channel = SensorChannel(self, chan_name)
            channels.append(channel)
            self.add_submodule(chan_name, channel)
        channels.lock()
        self.add_submodule("channels", channels)

        # add parameters
        self.add_parameter(
            "heater_output",
            get_cmd="HTR?",
            get_parser=float,
            label="heater output",
            unit="%",
        )

        self.add_parameter(
            "heater_range",
            get_cmd="RANGE?",
            get_parser=int,
            set_cmd="RANGE {}",
            val_mapping={"off": 0, "0.5W": 1, "5W": 2, "50W": 3},
            label="heater range",
        )

        self.add_parameter(
            "input",
            get_cmd=f"CSET? {self._loop}",
            set_cmd=f"CSET {self._loop},{{input}},1,1,1",
            get_parser=lambda ans: ans[0],
            label="input",
        )

        self.add_parameter(
            "setpoint",
            get_cmd="SETP? " + str(self._loop),
            set_cmd="SETP " + str(self._loop) + ", {}",
            get_parser=float,
            label="setpoint",
            unit="K",
        )

        self.add_parameter("P", label="P", parameter_class=GroupParameter)

        self.add_parameter("I", label="I", parameter_class=GroupParameter)

        self.add_parameter("D", label="D", parameter_class=GroupParameter)

        self.measure_group = Group(
            [self.P, self.I, self.D],
            get_cmd=f"PID? {self._loop}",
            set_cmd=f"PID {self._loop},{{P}},{{I}},{{D}}",
        )

        self.connect_message()



##Made by Dacen, copied over from cryocon_26 driver

from os.path import abspath, dirname
from time import sleep

from numpy import loadtxt
from qcodes import VisaInstrument
from qcodes.utils.validators import Enum, Numbers
from scipy.interpolate import interp1d

MAX_TEMP = 400  # K

dir = dirname(abspath(__file__))


class Cryocon_24(VisaInstrument):
    """Driver for the Cryo-con Model 24 temperature controller."""

    def __init__(self, name, address, terminator="\r\n", **kwargs):
        super().__init__(name, address, terminator=terminator, **kwargs)

        on_off_map = {True: "ON ", False: "OFF"}
        mapping_loop1 = {"high": "HI ", "medium": "MID", "low": "LOW"}
        mapping_loop2 = {"high": "HI ", "low": "LOW"}

        self.add_parameter(
            "control_enabled", get_cmd="control?", val_mapping=on_off_map
        )

        self.add_parameter(
            "chA_temperature",
            get_cmd=self._get_chA_temperature,
            get_parser=float,
            unit="K",
        )
        self.add_parameter(
            "chB_temperature",
            get_cmd=self._get_chB_temperature,
            get_parser=float,
            unit="K",
        )
        for loop in [1, 2]:
            l = f"loop{loop}_"
            self.add_parameter(
                l + "is_ramping",
                get_cmd=f"loop {loop}:ramp?",
                val_mapping=on_off_map,
                snapshot_value=False,
            )
        #             self.add_parameter(l + 'read_heater',
        #                                get_cmd='loop {}:htrread?',
        #                                get_parser=float,
        #                                snapshot_value=False,
        #                                unit='%')
        #             self.add_parameter(l + 'output_power',
        #                                get_cmd='loop {}:outp?',
        #                                get_parser=float,
        #                                snapshot_value=False,
        #                                unit='%')

        self.add_parameter(
            "loop1_load",
            get_cmd=self._get_loop1_load,
            set_cmd=self._set_loop1_load,
            get_parser=int,
            unit="Ohms",
            vals=Enum(25, 50),
        )

        if self.loop1_load() == 50:
            self.PID_TABLE_chA = loadtxt(
                dir + "/data/CyroCon24C_ChA_PID_Table_50Ohm.txt",
                skiprows=2,
                delimiter=",",
            )
        elif self.loop1_load() == 25:
            self.PID_TABLE_chA = loadtxt(
                dir + "/data/CyroCon24C_ChA_PID_Table_25Ohm.txt",
                skiprows=2,
                delimiter=",",
            )

        self.add_parameter(
            "chA_units",
            get_cmd="input A:units?",
            get_parser=str,
            set_cmd=self._set_chA_units,
            vals=Enum("K", "C", "F", "S"),
        )
        self.add_parameter(
            "chB_units",
            get_cmd="input B:units?",
            get_parser=str,
            set_cmd=self._set_chB_units,
            vals=Enum("K", "C", "F", "S"),
        )

        self.add_parameter(
            "loop1_source",
            get_cmd="loop 1:source?",
            get_parser=float,
            set_cmd=self._set_loop1_source,
            vals=Enum("A", "B"),
            snapshot_value=False,
        )
        self.add_parameter(
            "loop2_source",
            get_cmd="loop 2:source?",
            get_parser=float,
            set_cmd=self._set_loop1_source,
            vals=Enum("A", "B"),
            snapshot_value=False,
        )

        self.add_parameter(
            "loop1_setpoint",
            get_cmd=self._get_loop1_setpoint,
            get_parser=float,
            set_cmd=self._set_loop1_setpoint,
            vals=Numbers(0, MAX_TEMP),
        )
        self.add_parameter(
            "loop2_setpoint",
            get_cmd=self._get_loop2_setpoint,
            get_parser=float,
            set_cmd=self._set_loop2_setpoint,
            vals=Numbers(0, MAX_TEMP),
        )

        self.add_parameter(
            "loop1_type",
            get_cmd="loop 1:type?",
            set_cmd=self._set_loop1_type,
            val_mapping={
                "off": "OFF  ",
                "manual": "MAN  ",
                "PID": "PID  ",
            },
        )
        self.add_parameter(
            "loop2_type",
            get_cmd="loop 2:type?",
            set_cmd=self._set_loop2_type,
            val_mapping={
                "off": "OFF  ",
                "manual": "MAN  ",
                "PID": "PID  ",
            },
        )
        self.add_parameter(
            "loop1_range",
            get_cmd="loop 1:range?",
            set_cmd=self._set_loop1_range,
            val_mapping=mapping_loop1,
            snapshot_value=False,
        )
        self.add_parameter(
            "loop2_range",
            get_cmd="loop 2:range?",
            set_cmd=self._set_loop2_range,
            val_mapping=mapping_loop2,
            snapshot_value=False,
        )

        self.add_parameter(
            "loop1_P",
            get_cmd="loop 1:pgain?",
            set_cmd=self._set_loop1_P,
            get_parser=float,
            vals=Numbers(0, 1000),
        )
        self.add_parameter(
            "loop1_I",
            get_cmd="loop 1:igain?",
            get_parser=float,
            set_cmd=self._set_loop1_I,
            vals=Numbers(0, 1000),
        )
        self.add_parameter(
            "loop1_D",
            get_cmd="loop 1:dgain?",
            get_parser=float,
            set_cmd=self._set_loop1_D,
            vals=Numbers(0, 1000),
        )

        self.add_parameter(
            "loop2_P",
            get_cmd="loop 2:pgain?",
            get_parser=float,
            set_cmd=self._set_loop2_P,
            vals=Numbers(0, 1000),
        )
        self.add_parameter(
            "loop2_I",
            get_cmd="loop 2:igain?",
            get_parser=float,
            set_cmd=self._set_loop2_I,
            vals=Numbers(0, 1000),
        )
        self.add_parameter(
            "loop2_D",
            get_cmd="loop 2:dgain?",
            get_parser=float,
            set_cmd=self._set_loop2_D,
            vals=Numbers(0, 1000),
        )

        self.add_parameter(
            "loop1_manual_power",
            get_cmd="loop 1:pman?",
            set_cmd=self._set_loop1_manual_power,
            get_parser=float,
            vals=Numbers(0, 100),
            unit="%",
        )
        self.add_parameter(
            "loop2_manual_power",
            get_cmd="loop 2:pman?",
            set_cmd=self._set_loop2_manual_power,
            get_parser=float,
            vals=Numbers(0, 100),
            unit="%",
        )

        self.add_parameter(
            "sample_temperature",
            get_cmd=self._get_chA_temperature,
            set_cmd=self._set_sample_temperature,
            get_parser=float,
            unit="K",
        )

        self.add_parameter(
            "vti_temperature",
            get_cmd=self._get_chB_temperature,
            set_cmd=self._set_vti_temperature,
            get_parser=float,
            unit="K",
        )

        self.add_parameter(
            "simultaneous_temperature",
            set_cmd=self._set_simultaneous_temperature,
            get_cmd=self._get_chA_temperature,
            get_parser=float,
            unit="K",
        )

    def _execute(self, command):
        self.device_clear()
        self.write_raw(command)
        sleep(1.0)
        self.device_clear()

    def _ask(self, message):
        try:
            self.ask(message)
        except:
            sleep(1.0)
            self._ask(message)
        return self.ask(message)

    #     for channel in ['A', 'B']:
    #             c = 'ch{}_'.format(channel)
    #             self.add_parameter(c + 'temperature',
    #                                get_cmd='input? {}'.format(channel),
    #                                get_parser=float,
    #                               )
    def _get_chA_temperature(self):
        return self._ask("input? A")

    def _get_chB_temperature(self):
        return self._ask("input? B")

    def _set_chA_units(self, units):
        self._execute("input A:units " + units)

    def _set_chB_units(self, units):
        self._execute("input B:units " + units)

    def _set_loop1_load(self, load):
        # Set the PID table values depending on what the load is set to
        self._execute(f"loop 1:load {load}")
        if load == 50:
            self.PID_TABLE_chA = loadtxt(
                dir + "/data/CyroCon24C_ChA_PID_Table_50Ohm.txt",
                skiprows=2,
                delimiter=",",
            )
        elif load == 25:
            self.PID_TABLE_chA = loadtxt(
                dir + "/data/CyroCon24C_ChA_PID_Table_25Ohm.txt",
                skiprows=2,
                delimiter=",",
            )

    def _get_loop1_load(self):
        result = self._ask("loop 1:load?")
        return result

    def _set_loop1_source(self, channel):
        self._execute("loop 1:source " + channel)

    def _set_loop2_source(self, channel):
        self._execute("loop 2:source " + channel)

    def _get_loop1_setpoint(self):
        result = self.ask("loop 1:setpt?")
        return float(result[:-1])

    def _get_loop2_setpoint(self):
        result = self.ask("loop 2:setpt?")
        return float(result[:-1])

    def _set_loop1_setpoint(self, setpoint):
        self._execute(f"loop 1:setpt {setpoint}")

    def _set_loop2_setpoint(self, setpoint):
        self._execute(f"loop 2:setpt {setpoint}")

    def _set_loop1_type(self, type_):
        self._execute("loop 1:type " + type_)

    def _set_loop2_type(self, type_):
        self._execute("loop 2:type " + type_)

    def _set_loop1_range(self, range_):
        self._execute("loop 1:range " + range_)

    def _set_loop2_range(self, range_):
        self._execute("loop 2:range " + range_)

    def _set_loop1_P(self, P):
        self._execute(f"loop 1:pgain {P}")

    def _set_loop1_I(self, I):
        self._execute(f"loop 1:igain {I}")

    def _set_loop1_D(self, D):
        self._execute(f"loop 1:dgain {D}")

    def _set_loop2_P(self, P):
        self._execute(f"loop 2:pgain {P}")

    def _set_loop2_I(self, I):
        self._execute(f"loop 2:igain {I}")

    def _set_loop2_D(self, D):
        self._execute(f"loop 2:dgain {D}")

    def _set_loop1_manual_power(self, power):
        self._execute(f"loop 1:pman {power}")

    def _set_loop2_manual_power(self, power):
        self._execute(f"loop 2:pman {power}")

    def _set_sample_temperature(self, setpoint):
        check_set(self.loop1_setpoint, setpoint)
        self.update_sample_PID(setpoint)

    def _set_vti_temperature(self, setpoint):
        check_set(self.loop2_setpoint, setpoint)
        self.update_vti_PID(setpoint)

    def _set_simultaneous_temperature(self, setpoint):
        self._set_sample_temperature(setpoint)
        self._set_vti_temperature(setpoint)

    def update_sample_PID(self, setpoint):
        #         if self.loop1_load() == 50:
        #             P, I, D, heater_range = interpolate_PID_value(setpoint, PID_TABLE_chA_50Ohm, HEATER_RANGE_MAPPING_chA)
        #         elif self.loop1_load() == 25:
        #             P, I, D, heater_range = interpolate_PID_value(setpoint, PID_TABLE_chA_25Ohm, HEATER_RANGE_MAPPING_chA)
        P, I, D, heater_range = interpolate_PID_value(
            setpoint, self.PID_TABLE_chA, HEATER_RANGE_MAPPING_chA
        )
        check_set(self.loop1_P, P)
        check_set(self.loop1_I, I)
        check_set(self.loop1_D, D)
        check_set(self.loop1_range, heater_range)

    def update_vti_PID(self, setpoint):
        P, I, D, heater_range = interpolate_PID_value(
            setpoint, PID_TABLE_chB, HEATER_RANGE_MAPPING_chB
        )
        check_set(self.loop2_P, P)
        check_set(self.loop2_I, I)
        check_set(self.loop2_D, D)
        check_set(self.loop2_range, heater_range)

    def stop(self):
        self.write_raw("stop")
        sleep(2.0)
        self.device_clear()

    def control(self):
        self.write_raw("control")
        sleep(2.0)
        self.device_clear()


# PID_TABLE_chA_50Ohm = loadtxt('./Drivers/Cryocon/CyroCon24C_ChA_PID_Table_50Ohm.txt', skiprows = 2, delimiter = ',')
# PID_TABLE_chA_25Ohm = loadtxt('./Drivers/Cryocon/CyroCon24C_ChA_PID_Table_25Ohm.txt', skiprows = 2, delimiter = ',')
HEATER_RANGE_MAPPING_chA = {0: "low", 1: "medium", 2: "high"}

PID_TABLE_chB = loadtxt(
    dir + "/data/CyroCon24C_ChB_PID_Table.txt", skiprows=2, delimiter=","
)
HEATER_RANGE_MAPPING_chB = {0: "low", 1: "high"}


def interpolate_PID_value(T, PID_TABLE, HEATER_RANGE_MAPPING):
    # Interpolate the PID value for the temperature given in K
    P = float(interp1d(PID_TABLE[:, 0], PID_TABLE[:, 1])(T))
    I = float(interp1d(PID_TABLE[:, 0], PID_TABLE[:, 2])(T))
    D = float(interp1d(PID_TABLE[:, 0], PID_TABLE[:, 3])(T))
    heater_range = round(float(interp1d(PID_TABLE[:, 0], PID_TABLE[:, 4])(T)))
    return P, I, D, HEATER_RANGE_MAPPING[heater_range]


def check_set(param, setpoint):
    try:
        if param.get() == setpoint:
            return False
        else:
            param.set(setpoint)
            return True
    except:
        sleep(1.0)
        check_set(param, setpoint)



# heatmap_thread.py

import math
from collections import deque

import numpy as np
import pyqtgraph as pg
from PyQt5.QtCore import QObject, QTimer, pyqtSlot  # moved to visualization
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import (
    QCheckBox,
    QComboBox,
    QHBoxLayout,
    QLabel,
    QProgressBar,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

# Configure PyQtGraph for better performance
pg.setConfigOptions(
    antialias=False,  # Disable antialiasing for better performance
    useOpenGL=False,  # Don't use OpenGL - can cause issues
    crashWarning=True,  # Show warnings for debugging
)
pg.setConfigOption("background", "w")  # White background
pg.setConfigOption("foreground", "k")  # Black foreground


class Heatmap(QObject):
    """PyQtGraph-based heatmap plotter for Sweep2D.

    Provides high-performance real-time 2D visualization using PyQtGraph ImageView.
    Maintains the same API as the original matplotlib Heatmap class for compatibility.

    Attributes:
    ---------
    sweep:
        The parent sweep object.
    data_to_add:
        Queue to store heatmap data updates.
    param_surfaces:
        Dict mapping param_index -> surface dict with keys:
            'data' (2D np.array), 'dict' (nested dict of values), 'min', 'max'.
    figs_set:
        Changes to true after figures have been created.
    widget:
        Main QWidget containing the heatmap layout.
    layout_widget:
        PyQtGraph GraphicsLayoutWidget to host plot + colorbar.
    plot_item:
        PyQtGraph PlotItem used as the main view with axes.
    heatmap_data:
        2D numpy array containing the heatmap data.
    heatmap_dict:
        Dictionary mapping outer/inner parameter values to data.
    hist_item:
        HistogramLUTItem providing an interactive colorbar.
    cmap_box:
        QComboBox to pick colormap.
    auto_levels_chk:
        QCheckBox to toggle auto color scaling.
    param_box:
        QComboBox to pick which followed parameter to visualize.
    info_label:
        QLabel displaying current heatmap parameter.

    Methods:
    ---------
    create_figs()
        Creates the PyQtGraph heatmap figure for Sweep2D.
    add_data(data_dict)
        Adds data dictionary to be plotted on the heatmap.
    update_heatmap()
        Updates the heatmap with new data from the queue.
    clear()
        Closes all active figures.
    """

    def __init__(self, sweep, update_interval=200):
        """Initializes the PyQtGraph heatmap thread.

        Parameters
        ---------
        sweep:
            The parent sweep object.
        update_interval:
            Update interval in milliseconds (default 200ms = 5 FPS).
        """
        QObject.__init__(self)

        self.sweep = sweep
        self.data_to_add = deque([])
        self.param_surfaces = {}
        self.count = (
            0  # maintained for backward-compat, not relied upon when out_value provided
        )
        self.figs_set = False

        # PyQtGraph-specific attributes
        self.widget = None
        self.layout_widget = None
        self.plot_item = None
        self.image_item = None
        self.hist_item = None
        self.heatmap_data = None
        self.heatmap_dict = {}
        self.out_keys = []
        self.in_keys = []
        self.info_label = None
        self.progress_bar = None
        self.elapsed_label = None
        self.remaining_label = None

        # Timer for regular heatmap updates (5 FPS)
        self.update_timer = None
        self.update_interval = update_interval  # milliseconds
        self.needs_refresh = False
        self.max_items_per_update = 10  # Process max 10 items per timer tick

        # UI controls
        self.cmap_box = None
        self.auto_levels_chk = None
        self._auto_levels_enabled = True
        self.param_box = None
        self._param_indices = []  # Map combobox rows to in_sweep._params indices

    def handle_close(self, event):
        """Handle widget close event."""
        self.clear()
        event.accept()

    def handle_key_press(self, event):
        """Handle keyboard shortcuts for sweep control from heatmap window."""
        try:
            key = event.key()
            if key == pg.QtCore.Qt.Key_Escape:
                # Pause the 2D sweep gracefully
                try:
                    self.sweep.pause()
                except Exception:
                    pass
            elif key == pg.QtCore.Qt.Key_Return or key == pg.QtCore.Qt.Key_Enter:
                try:
                    self.sweep.resume()
                except Exception:
                    pass
            elif key == pg.QtCore.Qt.Key_Space:
                try:
                    # Flip the inner sweep direction
                    if (
                        hasattr(self.sweep, "in_sweep")
                        and self.sweep.in_sweep is not None
                    ):
                        self.sweep.in_sweep.flip_direction()
                except Exception:
                    pass
        except Exception:
            pass

    def create_figs(self):
        """Creates the PyQtGraph heatmap figure for the 2D sweep."""
        try:
            if self.figs_set:
                return

            # First, determine the resolution on each axis
            self.res_in = (
                math.ceil(
                    abs((self.sweep.in_stop - self.sweep.in_start) / self.sweep.in_step)
                )
                + 1
            )
            self.res_out = (
                math.ceil(
                    abs(
                        (self.sweep.out_stop - self.sweep.out_start)
                        / self.sweep.out_step
                    )
                )
                + 1
            )

            # Create the heatmap data matrix - initially as all 0s
            self.heatmap_dict = {}
            self.out_keys = []
            self.in_keys = []
            self.out_step = self.sweep.out_step
            self.in_step = self.sweep.in_step

            for x_out in np.linspace(
                self.sweep.out_start,
                self.sweep.out_stop,
                int(
                    abs(self.sweep.out_stop - self.sweep.out_start)
                    / abs(self.sweep.out_step)
                    + 1
                ),
                endpoint=True,
            ):
                self.heatmap_dict[x_out] = {}
                self.out_keys.append(x_out)
                for x_in in np.linspace(
                    self.sweep.in_start,
                    self.sweep.in_stop,
                    int(
                        abs(self.sweep.in_stop - self.sweep.in_start)
                        / abs(self.sweep.in_step)
                        + 1
                    ),
                    endpoint=True,
                ):
                    self.heatmap_dict[x_out][x_in] = 0
                    if x_in not in self.in_keys:
                        self.in_keys.append(x_in)

            self.out_keys = sorted(self.out_keys)
            self.in_keys = sorted(self.in_keys)

            # Initialize surfaces store and the current displayed surface
            self.param_surfaces = {}
            self.count = 0

            # Ensure a QApplication exists (safe in scripts and in notebooks once ensure_qt() has been called)
            try:
                pg.mkQApp()
            except Exception:
                pass

            # Create main widget and layout
            self.widget = QWidget()
            self.widget.setWindowTitle("MeasureIt - 2D Heatmap")
            self.widget.resize(800, 600)

            main_layout = QVBoxLayout(self.widget)

            # Add parameter info label
            plot_para = self.sweep.in_sweep._params[self.sweep.heatmap_ind]
            self.info_label = QLabel(
                f"2D Heatmap: {plot_para.label} ({plot_para.unit})"
            )
            self.info_label.setFont(QFont("Arial", 12))
            self.info_label.setStyleSheet(
                "QLabel { background-color: #f0f0f0; padding: 5px; }"
            )
            main_layout.addWidget(self.info_label)

            state = getattr(self.sweep, "progressState", None)
            if state is not None and getattr(state, "progress", None) is not None:
                progress_info = QHBoxLayout()
                progress_info.setContentsMargins(0, 0, 0, 0)
                progress_info.setSpacing(8)

                self.elapsed_label = QLabel("Elapsed: --")
                self.remaining_label = QLabel("Remaining: --")

                progress_info.addWidget(self.elapsed_label)
                progress_info.addStretch(1)
                progress_info.addWidget(self.remaining_label)

                self.progress_bar = QProgressBar()
                self.progress_bar.setRange(0, 1000)
                self.progress_bar.setValue(0)
                self.progress_bar.setTextVisible(True)

                main_layout.addLayout(progress_info)
                main_layout.addWidget(self.progress_bar)
                self._update_progress_widgets()

            # Controls row (colormap picker, auto-levels, reset)
            controls = QHBoxLayout()
            controls.setContentsMargins(0, 0, 0, 0)
            controls.setSpacing(8)
            controls.addWidget(QLabel("Colormap:"))
            self.cmap_box = QComboBox()
            # Prefer common scientific colormaps; filter by availability
            available_maps = []
            try:
                available_maps = list(pg.colormap.listMaps())
            except Exception:
                available_maps = []
            preferred = [
                "viridis",
                "plasma",
                "inferno",
                "magma",
                "cividis",
                "turbo",
                "gray",
                "grays",
                "jet",
                "hot",
                "coolwarm",
            ]
            cmaps = (
                [m for m in preferred if m in available_maps]
                or available_maps
                or ["viridis"]
            )
            for m in cmaps:
                self.cmap_box.addItem(m)
            self.cmap_box.setCurrentText("viridis" if "viridis" in cmaps else cmaps[0])
            self.cmap_box.currentTextChanged.connect(self._apply_colormap)
            controls.addWidget(self.cmap_box)

            self.auto_levels_chk = QCheckBox("Auto levels")
            self.auto_levels_chk.setChecked(True)
            self.auto_levels_chk.toggled.connect(self._toggle_auto_levels)
            controls.addWidget(self.auto_levels_chk)

            reset_btn = QPushButton("Reset view")
            reset_btn.clicked.connect(self._reset_view)
            controls.addWidget(reset_btn)

            # Parameter selector (followed params, excluding outer set_param)
            controls.addWidget(QLabel("Signal:"))
            self.param_box = QComboBox()
            self._rebuild_param_box()
            self.param_box.currentIndexChanged.connect(self._on_param_selected)
            controls.addWidget(self.param_box)

            controls.addStretch(1)
            main_layout.addLayout(controls)

            # Create GraphicsLayout with a PlotItem (left) and HistogramLUTItem (right)
            self.layout_widget = pg.GraphicsLayoutWidget()
            main_layout.addWidget(self.layout_widget)

            self.plot_item = self.layout_widget.addPlot(row=0, col=0)
            self.plot_item.setLabel(
                "bottom", f"{self.sweep.in_param.label}", units=self.sweep.in_param.unit
            )
            self.plot_item.setLabel(
                "left", f"{self.sweep.set_param.label}", units=self.sweep.set_param.unit
            )
            self.plot_item.showGrid(x=True, y=True, alpha=0.3)

            # Create ImageItem for the heatmap data
            self.image_item = pg.ImageItem()
            self.plot_item.addItem(self.image_item)

            # Add interactive colorbar/histogram
            try:
                self.hist_item = pg.HistogramLUTItem()
                self.hist_item.setImageItem(self.image_item)
                self.layout_widget.addItem(self.hist_item, row=0, col=1)
            except Exception:
                self.hist_item = None  # fallback silently if unavailable

            # Store coordinate transform parameters for updates
            self.pos = [self.sweep.in_start, self.sweep.out_start]
            self.scale = [
                (self.sweep.in_stop - self.sweep.in_start) / self.res_in,
                (self.sweep.out_stop - self.sweep.out_start) / self.res_out,
            ]

            # Ensure an initial surface for the currently selected parameter
            self._ensure_surface(self.sweep.heatmap_ind)
            self.heatmap_data = self.param_surfaces[self.sweep.heatmap_ind]["data"]
            # Set up the initial image with proper scaling
            self.image_item.setImage(self.heatmap_data)

            # Set the coordinate transformation for proper axis scaling
            # Use setRect as the primary method (more reliable than transform)
            try:
                from PyQt5.QtCore import QRectF
                # setRect takes (x, y, width, height) in data coordinates
                self.image_item.setRect(
                    QRectF(
                        self.sweep.in_start,
                        self.sweep.out_start,
                        self.sweep.in_stop - self.sweep.in_start,   # Total width
                        self.sweep.out_stop - self.sweep.out_start, # Total height
                    )
                )
            except Exception:
                # Fallback: use transform if setRect fails
                try:
                    transform = pg.QtGui.QTransform()
                    transform.translate(self.pos[0], self.pos[1])
                    # Use (N-1) for correct pixel spacing
                    transform.scale(
                        (self.sweep.in_stop - self.sweep.in_start) / (self.res_in - 1),
                        (self.sweep.out_stop - self.sweep.out_start) / (self.res_out - 1)
                    )
                    self.image_item.setTransform(transform)
                except Exception:
                    pass

            # Connect close and keyboard events
            self.widget.closeEvent = self.handle_close
            self.widget.keyPressEvent = self.handle_key_press
            self.widget.setFocusPolicy(pg.QtCore.Qt.StrongFocus)

            # Show the widget
            self.widget.show()

            # Start the update timer for regular heatmap refreshes
            self.update_timer = QTimer()
            self.update_timer.timeout.connect(self.update_heatmap)
            self.update_timer.start(self.update_interval)

            # Apply initial colormap
            self._apply_colormap(
                self.cmap_box.currentText() if self.cmap_box is not None else "viridis"
            )

            self.figs_set = True

            self._update_progress_widgets()

        except Exception as e:
            print(f"Error creating heatmap figures: {e}")
            import traceback

            traceback.print_exc()
            self.figs_set = False

    @pyqtSlot(object)
    def add_data(self, data_dict):
        """Feeds the thread data dictionary to add to the heatmap.

        Parameters
        ---------
        data_dict:
            Dictionary containing 'forward' and 'backward' data tuples (x_data, y_data)
            from the plotter thread.
        """
        try:
            if data_dict is None:
                self.data_to_add.append(None)
            else:
                # Accept legacy payloads (only 'forward'/'backward') or new ones with 'param_index' and 'out_value'
                if "param_index" not in data_dict:
                    data_dict = dict(data_dict)  # shallow copy
                    data_dict["param_index"] = getattr(self.sweep, "heatmap_ind", 0)
                if "out_value" not in data_dict:
                    data_dict["out_value"] = getattr(self.sweep, "out_setpoint", None)
                self.data_to_add.append(data_dict)
        except Exception as e:
            print(f"Error in heatmap add_data: {e}")
            import traceback

            traceback.print_exc()

    def add_to_heatmap(self, data_dict):
        """Processes data dictionary for one complete inner sweep and updates heatmap data.

        Parameters
        ---------
        data_dict:
            Dictionary with 'forward' and 'backward' data tuples (x_data, y_data).
        """
        if "forward" not in data_dict:
            return

        # Determine target param and out value
        param_index = data_dict.get(
            "param_index", getattr(self.sweep, "heatmap_ind", 0)
        )
        out_value = data_dict.get(
            "out_value", getattr(self.sweep, "out_setpoint", None)
        )
        self._ensure_surface(param_index)

        # Get the complete inner sweep data
        x_data, y_data = data_dict["forward"]

        if len(x_data) == 0 or len(y_data) == 0:
            return

        # Remove NaN values
        valid_mask = ~(np.isnan(x_data) | np.isnan(y_data))
        x_clean = x_data[valid_mask]
        y_clean = y_data[valid_mask]

        if len(x_clean) == 0:
            return

        # Determine row based on provided out_value if available; otherwise fall back to current count index
        if out_value is not None:
            current_out_key = min(self.out_keys, key=lambda k: abs(k - out_value))
            out_index = self.out_keys.index(current_out_key)
        else:
            if self.count >= len(self.out_keys):
                print(
                    f"Warning: heatmap count {self.count} exceeds out_keys length {len(self.out_keys)}"
                )
                return
            current_out_key = self.out_keys[self.count]
            out_index = self.count

        # Map inner sweep data to heatmap grid
        for x_val, y_val in zip(x_clean, y_clean):
            # Find closest in_key for this x value
            closest_in_key = min(self.in_keys, key=lambda k: abs(k - x_val))
            self.param_surfaces[param_index]["dict"][current_out_key][
                closest_in_key
            ] = y_val

            # Track min/max values per param
            if y_val > self.param_surfaces[param_index]["max"]:
                self.param_surfaces[param_index]["max"] = y_val
            if y_val < self.param_surfaces[param_index]["min"]:
                self.param_surfaces[param_index]["min"] = y_val

        # Update the data array row
        row_idx = self.res_out - out_index - 1  # Flip for proper orientation
        for i, in_key in enumerate(self.in_keys):
            if i < self.res_in:
                self.param_surfaces[param_index]["data"][row_idx][i] = (
                    self.param_surfaces[param_index]["dict"][current_out_key][in_key]
                )

        # Mark that this param surface has new data
        if not hasattr(self, "_dirty_params"):
            self._dirty_params = set()
        self._dirty_params.add(param_index)
        self.needs_refresh = True

    def update_heatmap(self):
        """Updates the heatmap with new data from the queue using batch processing."""
        if not self.figs_set:
            return

        # Process limited batch of data to prevent UI blocking
        items_processed = 0
        while len(self.data_to_add) > 0 and items_processed < self.max_items_per_update:
            data_dict = self.data_to_add.popleft()
            if data_dict is None:
                break
            self.add_to_heatmap(data_dict)
            items_processed += 1
            self.needs_refresh = True

        self._update_progress_widgets()
        # Update display with proper scaling while preserving view
        if self.image_item is not None and (self.needs_refresh or items_processed > 0):
            # Remember current view state
            view_box = self.plot_item.getViewBox()
            view_range = view_box.viewRange()

            # Update the image data
            # Ensure our current display surface points to the selected param
            current_idx = getattr(self.sweep, "heatmap_ind", 0)
            self._ensure_surface(current_idx)
            self.heatmap_data = self.param_surfaces[current_idx]["data"]
            self.image_item.setImage(self.heatmap_data)

            # Set color levels: auto or leave to user via histogram LUT
            if self._auto_levels_enabled:
                try:
                    # Robust percentiles if data available; fallback to min/max trackers
                    finite = self.heatmap_data[np.isfinite(self.heatmap_data)]
                    if finite.size > 0:
                        lo, hi = np.nanpercentile(finite, [1, 99])
                        if hi > lo:
                            if self.hist_item is not None:
                                self.hist_item.setLevels(lo, hi)
                            else:
                                self.image_item.setLevels([lo, hi])
                        else:
                            stats = self.param_surfaces[current_idx]
                            if self.hist_item is not None:
                                self.hist_item.setLevels(stats["min"], stats["max"])
                            else:
                                self.image_item.setLevels([stats["min"], stats["max"]])
                    else:
                        stats = self.param_surfaces[current_idx]
                        if self.hist_item is not None:
                            self.hist_item.setLevels(stats["min"], stats["max"])
                        else:
                            self.image_item.setLevels([stats["min"], stats["max"]])
                except Exception:
                    # Last resort: use tracked min/max
                    stats = self.param_surfaces[current_idx]
                    if self.hist_item is not None:
                        self.hist_item.setLevels(stats["min"], stats["max"])
                    else:
                        self.image_item.setLevels([stats["min"], stats["max"]])

            # Restore view range to preserve user zoom/pan
            view_box.setRange(xRange=view_range[0], yRange=view_range[1], padding=0)

            self.needs_refresh = False

    def _rebuild_param_box(self):
        """Populate the parameter selector with followed measurement parameters."""
        try:
            if self.param_box is None:
                return
            self.param_box.blockSignals(True)
            self.param_box.clear()
            self._param_indices = []
            # Choose candidates: prefer user-provided list on the sweep
            indices = getattr(self.sweep, "heatmap_param_indices", None)
            candidates = []
            if isinstance(indices, list) and len(indices) > 0:
                for idx in indices:
                    if 0 <= idx < len(self.sweep.in_sweep._params):
                        p = self.sweep.in_sweep._params[idx]
                        if p is not self.sweep.set_param:
                            candidates.append((idx, p))
            # Fallback to all followed params except outer set_param
            if len(candidates) == 0:
                for idx, p in enumerate(self.sweep.in_sweep._params):
                    if p is not self.sweep.set_param:
                        candidates.append((idx, p))
            for idx, p in candidates:
                lab = p.label if getattr(p, "label", None) else p.name
                txt = f"{lab} ({p.unit})" if getattr(p, "unit", None) else lab
                self.param_box.addItem(txt)
                self._param_indices.append(idx)
            # Set current selection to match sweep.heatmap_ind
            try:
                current = self._param_indices.index(self.sweep.heatmap_ind)
            except ValueError:
                current = 0 if self._param_indices else -1
            if current >= 0:
                self.param_box.setCurrentIndex(current)
            self.param_box.blockSignals(False)
        except Exception:
            self.param_box.blockSignals(False)

    @pyqtSlot()
    def refresh_param_list(self):
        """Public slot to rebuild parameter selector from Sweep2D configuration."""
        try:
            self._rebuild_param_box()
            # Also update info label text in case the current param changed
            try:
                p = self.sweep.in_sweep._params[self.sweep.heatmap_ind]
                lab = p.label if getattr(p, "label", None) else p.name
                txt = f"2D Heatmap: {lab} ({p.unit})"
                if self.info_label is not None:
                    self.info_label.setText(txt)
            except Exception:
                pass
        except Exception:
            pass

    def _on_param_selected(self, combo_index: int):
        """Handle selection change: update sweep heatmap index and reset content."""
        try:
            if combo_index < 0 or combo_index >= len(self._param_indices):
                return
            real_index = self._param_indices[combo_index]
            # Update Sweep2D selection using existing helper
            try:
                target_param = self.sweep.in_sweep._params[real_index]
                self.sweep.follow_heatmap_param(target_param)
            except Exception:
                self.sweep.heatmap_ind = real_index
            # Update label and reset displayed data
            try:
                p = self.sweep.in_sweep._params[self.sweep.heatmap_ind]
                lab = p.label if getattr(p, "label", None) else p.name
                txt = f"2D Heatmap: {lab} ({p.unit})"
                if self.info_label is not None:
                    self.info_label.setText(txt)
            except Exception:
                pass
            # Switch the displayed surface without clearing accumulated data
            self._ensure_surface(self.sweep.heatmap_ind)
            self.heatmap_data = self.param_surfaces[self.sweep.heatmap_ind]["data"]
            # Refresh the image to show the newly selected surface
            view_box = self.plot_item.getViewBox()
            view_range = view_box.viewRange()
            self.image_item.setImage(self.heatmap_data)

            # Re-apply coordinate transformation after setImage (critical for correct axis scaling)
            try:
                from PyQt5.QtCore import QRectF
                self.image_item.setRect(
                    QRectF(
                        self.sweep.in_start,
                        self.sweep.out_start,
                        self.sweep.in_stop - self.sweep.in_start,
                        self.sweep.out_stop - self.sweep.out_start,
                    )
                )
            except Exception:
                # Fallback: use transform if setRect fails
                try:
                    transform = pg.QtGui.QTransform()
                    transform.translate(self.pos[0], self.pos[1])
                    transform.scale(
                        (self.sweep.in_stop - self.sweep.in_start) / (self.res_in - 1),
                        (self.sweep.out_stop - self.sweep.out_start) / (self.res_out - 1)
                    )
                    self.image_item.setTransform(transform)
                except Exception:
                    pass

            # Update color levels for the newly selected parameter
            if self._auto_levels_enabled:
                try:
                    current_idx = self.sweep.heatmap_ind
                    finite = self.heatmap_data[np.isfinite(self.heatmap_data)]
                    if finite.size > 0:
                        lo, hi = np.nanpercentile(finite, [1, 99])
                        if hi > lo:
                            if self.hist_item is not None:
                                self.hist_item.setLevels(lo, hi)
                            else:
                                self.image_item.setLevels([lo, hi])
                        else:
                            stats = self.param_surfaces[current_idx]
                            if self.hist_item is not None:
                                self.hist_item.setLevels(stats["min"], stats["max"])
                            else:
                                self.image_item.setLevels([stats["min"], stats["max"]])
                    else:
                        stats = self.param_surfaces[current_idx]
                        if self.hist_item is not None:
                            self.hist_item.setLevels(stats["min"], stats["max"])
                        else:
                            self.image_item.setLevels([stats["min"], stats["max"]])
                except Exception:
                    pass

            # Restore view
            view_box.setRange(xRange=view_range[0], yRange=view_range[1], padding=0)

            # Mark as needing refresh to ensure display updates on Windows
            self.needs_refresh = True
        except Exception:
            pass

    def _reset_content(self):
        """Reset incoming queue and counters; keep accumulated surfaces."""
        try:
            self.data_to_add.clear()
            # Do not zero stored data; only reset per-row fallback counter
            self.count = 0
            self.needs_refresh = False
        except Exception:
            pass

    def _ensure_surface(self, param_index: int):
        """Ensure that a surface store exists for the given parameter index."""
        if param_index not in self.param_surfaces:
            # Create fresh 2D array and dict for this param
            arr = np.zeros((self.res_out, self.res_in))
            val_dict = {}
            for x_out in self.out_keys:
                val_dict[x_out] = {}
                for x_in in self.in_keys:
                    val_dict[x_out][x_in] = 0
            self.param_surfaces[param_index] = {
                "data": arr,
                "dict": val_dict,
                "min": float("inf"),
                "max": float("-inf"),
            }

    def _apply_colormap(self, name: str):
        """Apply selected colormap to image and histogram."""
        try:
            cmap = None
            try:
                cmap = pg.colormap.get(name)
            except Exception:
                pass

            if cmap is not None:
                # Apply to histogram if available (preferred)
                if self.hist_item is not None and hasattr(
                    self.hist_item, "setColorMap"
                ):
                    self.hist_item.setColorMap(cmap)
                # Also apply LUT to the image for consistency
                try:
                    lut = cmap.getLookupTable(alpha=False)
                    self.image_item.setLookupTable(lut)
                except Exception:
                    pass
            else:
                # Fallback: simple grayscale LUT
                lut = np.stack([np.linspace(0, 255, 256)] * 3, axis=1)
                self.image_item.setLookupTable(lut)
        except Exception:
            pass

    def _toggle_auto_levels(self, checked: bool):
        self._auto_levels_enabled = bool(checked)

    def _update_progress_widgets(self):
        if self.progress_bar is None:
            return

        state = getattr(self.sweep, "progressState", None)
        if state is None:
            return

        progress = getattr(state, "progress", None)
        if progress is None:
            self.progress_bar.setFormat("Progress: --")
            self.progress_bar.setValue(0)
        else:
            self.progress_bar.setFormat("Progress: %p%")
            progress_value = int(max(0.0, min(1.0, progress)) * 1000)
            self.progress_bar.setValue(progress_value)

        def _format_seconds(value):
            if value is None:
                return "--"
            try:
                if math.isinf(value):
                    return "∞"
            except TypeError:
                return "--"
            return f"{max(0.0, value):.1f} s"

        if self.elapsed_label is not None:
            self.elapsed_label.setText(
                f"Elapsed: {_format_seconds(state.time_elapsed)}"
            )
        if self.remaining_label is not None:
            self.remaining_label.setText(
                f"Remaining: {_format_seconds(state.time_remaining)}"
            )

    def _reset_view(self):
        try:
            if self.plot_item is not None:
                self.plot_item.enableAutoRange(x=True, y=True)
        except Exception:
            pass

    def clear(self):
        """Closes all active figures."""
        if self.figs_set:
            # Stop the update timer
            if self.update_timer is not None:
                self.update_timer.stop()
                self.update_timer = None

            self.figs_set = False
            self.needs_refresh = False
            if self.widget is not None:
                self.widget.close()
                self.widget = None
                self.layout_widget = None
                self.plot_item = None
                self.image_item = None
                self.hist_item = None
                self.heatmap_data = None
                self.heatmap_dict = {}
                self.out_keys = []
                self.in_keys = []



from qcodes.instrument.ip import IPInstrument


class SCM10(IPInstrument):
    def __init__(
        self,
        name,
        address="192.168.1.4",
        port=9760,
        timeout=5,
        terminator="\r\n",
        *args,
        **kwargs,
    ):
        super().__init__(name, address, port, timeout, terminator, *args, **kwargs)

        self.add_parameter(
            "temperature",
            label="temperature",
            unit="K",
            get_cmd="T?",
            get_parser=lambda resp: float(resp.replace("T ", "")),
        )

        self.add_parameter(
            "sensor_reading",
            label="sensor reading",
            unit="V",
            get_cmd="SDAT?",
            get_parser=lambda resp: resp.replace("SDAT ", ""),
        )

        self.add_parameter(
            "error_status",
            label="error status",
            get_cmd="ERSTA?",
            get_parser=lambda resp: resp.replace("ERSTA ", ""),
        )

        self.connect_message()



# This Python file uses the following encoding: utf-8
# Etienne Dumur <etienne.dumur@gmail.com>, September 2020
# Jiaqi Cai <jiaqic@mit.edu>, Oct., 2025
# HTTP API rewrite inspired by BlueFTC (https://github.com/eliasankerhold/BlueFTC)

import logging
from functools import partial
from random import randint
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
import requests
from qcodes.instrument.base import Instrument
from qcodes.utils.validators import Bool, Numbers
from urllib3 import disable_warnings
from urllib3.exceptions import InsecureRequestWarning
from warnings import warn


class PIDConfigException(Exception):
    """Raised when PID configuration setup fails."""


class APIError(Exception):
    """Wrap errors returned by the BlueFors HTTP API."""

    def __init__(self, payload: Dict[str, Any]) -> None:
        error_block = payload.get("error", {})
        details = error_block.get("details")
        if isinstance(details, list) and details:
            messages = [
                f"Code: {entry.get('code')}, Reason: {entry.get('name')}"
                for entry in details
                if isinstance(entry, dict)
            ]
        elif error_block:
            messages = [
                f"Code: {error_block.get('code')}, Reason: {error_block.get('name')}"
            ]
        else:
            messages = ["Unknown BlueFors controller error."]
        description = error_block.get("description", "")
        message = (
            f"{error_block.get('name', 'BlueFors API error')}: "
            f"{description} {' | '.join(messages)}"
        ).strip()
        super().__init__(message)
        self.payload: Dict[str, Any] = payload


class _BlueForsAPIClient:
    """
    Helper that talks to the BlueFors Controller HTTP API.

    The implementation follows the reference BlueFTC driver but is trimmed for
    integration inside a QCoDeS instrument.
    """

    def __init__(
        self,
        host: str,
        api_key: str,
        *,
        port: int = 49098,
        mixing_chamber_channel_id: Optional[int] = None,
        mixing_chamber_heater_id: Optional[int] = None,
        pid_calibration_path: Optional[str] = None,
        enable_maxigauge: bool = False,
        emulate: bool = False,
        verify_ssl: bool = False,
        timeout: float = 5.0,
        auto_pid_calibration: bool = True,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        if not host:
            raise PIDConfigException("A valid host/IP address is required.")
        if not api_key:
            raise PIDConfigException(
                "An API key is required to access the BlueFors controller."
            )

        self._base_url = f"https://{host}:{port}"
        self._key = api_key
        self._timeout = timeout
        self._verify_ssl = verify_ssl
        self._session = requests.Session()
        self._maxigauge_pressure = enable_maxigauge
        self._emulate = emulate
        self._auto_pid_calibration = auto_pid_calibration
        self._log = logger or logging.getLogger(__name__)
        self._mixing_chamber_channel_id = mixing_chamber_channel_id
        self._mixing_chamber_heater = (
            f"driver.bftc.data.heaters.heater_{mixing_chamber_heater_id}"
            if mixing_chamber_heater_id is not None
            else None
        )
        self._has_mxc = mixing_chamber_channel_id is not None

        if not verify_ssl:
            disable_warnings(InsecureRequestWarning)

        self._valid_pid_config = False
        self._pid_calib_setpoints: Optional[np.ndarray] = None
        self._pid_calib_pid: Optional[np.ndarray] = None
        if pid_calibration_path is not None:
            self._load_pid_config(pid_calibration_path)

    @staticmethod
    def _device_key(device: str, target: str) -> str:
        return f"{device}.{target}"

    def _value_url(self, device: str, target: str) -> str:
        path = device.replace(".", "/")
        return (
            f"{self._base_url}/values/{path}/{target}/"
            f"?prettyprint=1&key={self._key}"
        )

    def _post_url(self) -> str:
        return f"{self._base_url}/values/?prettyprint=1&key={self._key}"

    def _nan_payload(
        self, device: str, target: str, status: str = "ERROR"
    ) -> Dict[str, Any]:
        key = self._device_key(device, target)
        return {
            "data": {
                key: {
                    "content": {
                        "latest_valid_value": {"value": float("nan"), "status": status}
                    }
                }
            }
        }

    def _mock_value(self, device: str, target: str) -> Dict[str, Any]:
        key = self._device_key(device, target)
        value = randint(1, 100)
        if target in ("active", "pid_mode"):
            mock_value: Any = "1" if value % 2 else "0"
        else:
            mock_value = float(value)
        return {
            "data": {
                key: {
                    "content": {
                        "latest_valid_value": {"value": mock_value, "status": "SYNCHRONIZED"}
                    }
                }
            }
        }

    def _get_value_request(self, device: str, target: str) -> Dict[str, Any]:
        if self._emulate:
            self._log.debug("EMULATE GET %s.%s", device, target)
            return self._mock_value(device, target)

        url = self._value_url(device, target)
        self._log.debug("GET %s", url)
        try:
            response = self._session.get(
                url,
                timeout=self._timeout,
                verify=self._verify_ssl,
            )
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as err:
            self._log.error("BlueFors GET %s failed: %s", url, err)
            return self._nan_payload(device, target)

    def _set_value_request(self, device: str, target: str, value: Any) -> None:
        if self._emulate:
            self._log.debug("EMULATE POST %s.%s = %s", device, target, value)
            return

        body = {
            "data": {
                self._device_key(device, target): {"content": {"value": value}}
            }
        }
        url = self._post_url()
        self._log.debug("POST %s payload=%s", url, body)
        response = self._session.post(
            url,
            json=body,
            timeout=self._timeout,
            verify=self._verify_ssl,
        )
        response.raise_for_status()

    def _apply_values_request(self, device: str) -> None:
        if self._emulate:
            self._log.debug("EMULATE APPLY %s", device)
            return

        body = {
            "data": {
                self._device_key(device, "write"): {"content": {"call": 1}}
            }
        }
        url = self._post_url()
        self._log.debug("POST %s payload=%s", url, body)
        response = self._session.post(
            url,
            json=body,
            timeout=self._timeout,
            verify=self._verify_ssl,
        )
        response.raise_for_status()

    def _get_synchronization_status(
        self, data: Dict[str, Any], device: str, target: str
    ) -> str:
        key = self._device_key(device, target)
        try:
            return data["data"][key]["content"]["latest_valid_value"]["status"]
        except KeyError:
            self._log.warning(
                "Could not verify synchronization status for %s.%s",
                device,
                target,
            )
            return "INVALID"

    @staticmethod
    def _handle_status_response(
        status: str, target: str, set_operation: bool = False
    ) -> int:
        info = " while setting value" if set_operation else ""
        if status == "INVALID":
            print(f"Warning{info}: The target '{target}' is invalid.")
            return 0
        if status in ("CHANGED", "SYNCHRONIZED", "INDEPENDENT"):
            return 1
        if status == "DISCONNECTED":
            print(
                f"Warning{info}: Target '{target}' is disconnected on the controller."
            )
            return 0
        if status == "QUEUED":
            print(
                f"Warning{info}: Target '{target}' is queued; verify synchronization."
            )
            return 2
        print(f"Warning{info}: Received invalid status '{status}' for '{target}'.")
        return 0

    def _get_value_from_response(
        self, data: Dict[str, Any], device: str, target: str
    ) -> Any:
        status = self._get_synchronization_status(data, device, target)
        self._handle_status_response(status=status, target=target)
        key = self._device_key(device, target)
        return data["data"][key]["content"]["latest_valid_value"]["value"]

    @staticmethod
    def _device_for_channel(channel: int) -> str:
        return f"mapper.heater_mappings_bftc.device.c{channel}"

    def get_channel_data(self, channel: int, target: str) -> Any:
        device = self._device_for_channel(channel)
        data = self._get_value_request(device, target)
        try:
            return self._get_value_from_response(data, device, target)
        except KeyError as err:
            raise APIError(data) from err

    def get_channel_temperature(self, channel: int) -> float:
        return float(self.get_channel_data(channel, "temperature"))

    def get_channel_resistance(self, channel: int) -> float:
        return float(self.get_channel_data(channel, "resistance"))

    def _ensure_mxc_channel(self) -> int:
        if not self._has_mxc or self._mixing_chamber_channel_id is None:
            raise PIDConfigException("Mixing chamber channel ID not configured.")
        return int(self._mixing_chamber_channel_id)

    def _ensure_mxc_heater(self) -> str:
        if self._mixing_chamber_heater is None:
            raise PIDConfigException("Mixing chamber heater ID not configured.")
        return self._mixing_chamber_heater

    def get_mxc_temperature(self) -> float:
        channel = self._ensure_mxc_channel()
        return self.get_channel_temperature(channel)

    def get_mxc_resistance(self) -> float:
        channel = self._ensure_mxc_channel()
        return self.get_channel_resistance(channel)

    def get_mxc_heater_value(self, target: str) -> Any:
        heater = self._ensure_mxc_heater()
        data = self._get_value_request(heater, target)
        try:
            return self._get_value_from_response(data, heater, target)
        except KeyError as err:
            raise APIError(data) from err

    def check_heater_value_synced(self, target: str) -> bool:
        heater = self._ensure_mxc_heater()
        data = self._get_value_request(heater, target)
        status = self._get_synchronization_status(data, heater, target)
        return bool(
            self._handle_status_response(status, target=target, set_operation=True)
        )

    def set_mxc_heater_value(self, target: str, value: Any) -> bool:
        heater = self._ensure_mxc_heater()
        self._log.info("Mixing chamber heater: setting %s to %s", target, value)
        self._set_value_request(heater, target, value)
        self._apply_values_request(heater)
        synced = self.check_heater_value_synced(target)
        if synced:
            self._log.info("Mixing chamber heater: %s synced", target)
        return synced

    def get_mxc_heater_status(self) -> bool:
        return self.get_mxc_heater_value("active") == "1"

    def set_mxc_heater_status(self, enabled: bool) -> bool:
        new_value = "1" if enabled else "0"
        return self.set_mxc_heater_value("active", new_value)

    def toggle_mxc_heater(self, status: str) -> bool:
        if status not in {"on", "off"}:
            raise PIDConfigException("Status must be 'on' or 'off'.")
        return self.set_mxc_heater_status(status == "on")

    def get_mxc_heater_power(self) -> float:
        return float(self.get_mxc_heater_value("power")) * 1e6

    def set_mxc_heater_power(self, power_uW: float) -> bool:
        if power_uW < 0 or power_uW > 5000:
            raise PIDConfigException(
                "Power should be in the range of 0 to 5000 microwatts."
            )
        return self.set_mxc_heater_value("power", power_uW / 1e6)

    def get_mxc_heater_setpoint(self) -> float:
        return float(self.get_mxc_heater_value("setpoint"))

    def set_mxc_heater_setpoint(
        self, temperature_mK: float, use_pid_calib: Optional[bool] = None
    ) -> bool:
        if temperature_mK >= 1e3:
            raise PIDConfigException(
                f"Mixing chamber setpoint cannot be over 1 K. Requested {temperature_mK} mK."
            )

        apply_pid = (
            self._auto_pid_calibration if use_pid_calib is None else use_pid_calib
        )
        if apply_pid and self._valid_pid_config:
            if self._pid_calib_setpoints is None or self._pid_calib_pid is None:
                warn("PID calibration table not initialized.")
            else:
                temperature_k = temperature_mK * 1e-3
                closest = int(
                    np.argmin(np.abs(self._pid_calib_setpoints - temperature_k))
                )
                target_mk = self._pid_calib_setpoints[closest] * 1e3
                self._log.info(
                    "Using PID calibration closest to %.3f mK.", target_mk
                )
                self.set_mxc_heater_pid_config(*self._pid_calib_pid[closest])
        elif apply_pid and not self._valid_pid_config:
            warn(
                "PID calibration requested but no valid table is available; "
                "using current PID values."
            )

        return self.set_mxc_heater_value("setpoint", temperature_mK)

    def get_mxc_heater_mode(self) -> bool:
        return self.get_mxc_heater_value("pid_mode") == "1"

    def set_mxc_heater_mode(self, enable: bool) -> bool:
        value = "1" if enable else "0"
        return self.set_mxc_heater_value("pid_mode", value)

    def get_mxc_heater_pid_config(self) -> List[float]:
        pid_values: List[float] = []
        for axis in ("p", "i", "d"):
            pid_values.append(float(self.get_mxc_heater_value(f"pid_{axis}")))
        return pid_values

    def set_mxc_heater_pid_config(
        self, p: Optional[float] = None, i: Optional[float] = None, d: Optional[float] = None
    ) -> bool:
        success = True
        for value, axis in zip((p, i, d), ("p", "i", "d")):
            if value is not None:
                success &= self.set_mxc_heater_value(f"pid_{axis}", value)
        return success

    def get_maxigauge_channel(self, channel: int) -> float:
        if not self._maxigauge_pressure:
            raise PIDConfigException(
                "Activate Maxigauge reading to query pressure values."
            )
        device = "driver.maxigauge.pressures"
        target = f"p{channel}"
        data = self._get_value_request(device, target)
        try:
            value = self._get_value_from_response(data, device, target)
            return float(value) * 1e3
        except KeyError as err:
            raise APIError(data) from err

    def _load_pid_config(self, path: str) -> None:
        try:
            pid_config = np.loadtxt(path, dtype="float", delimiter=",", skiprows=1)
        except FileNotFoundError as err:
            warn(
                f"Encountered error while loading PID config '{path}': {err}.\n"
                "Continuing without automatic PID parameter adjustment."
            )
            return

        pid_config = np.atleast_2d(pid_config)
        if pid_config.shape[1] < 4:
            warn(
                "PID calibration file must contain at least four columns: "
                "[setpoint (mK), P, I, D]."
            )
            return

        setpoints = pid_config[:, 0]
        pid_values = pid_config[:, 1:4]
        sort_index = np.argsort(setpoints)
        self._pid_calib_setpoints = setpoints[sort_index] * 1e-3
        self._pid_calib_pid = pid_values[sort_index]
        self._valid_pid_config = True
        self._log.info(
            "PID calibration loaded from %s (%d entries).",
            path,
            len(self._pid_calib_setpoints),
        )


class BlueFors(Instrument):
    """
    QCoDeS driver for BlueFors fridges using the HTTP API exposed by
    the BlueFors Temperature Controller software.

    Temperatures, resistances, pressures and mixing chamber heater controls
    are queried from the controller directly, which avoids the latency of
    scraping log files on disk.
    """

    def __init__(
        self,
        name: str,
        folder_path: Optional[str],
        channel_vacuum_can: Optional[int],
        channel_pumping_line: Optional[int],
        channel_compressor_outlet: Optional[int],
        channel_compressor_inlet: Optional[int],
        channel_mixture_tank: Optional[int],
        channel_venting_line: Optional[int],
        channel_50k_plate: Optional[int],
        channel_4k_plate: Optional[int],
        channel_still: Optional[int],
        channel_mixing_chamber: Optional[int],
        channel_magnet: Optional[int] = None,
        channel_fse: Optional[int] = None,
        *,
        host: str,
        api_key: str,
        port: int = 49098,
        mixing_chamber_heater_id: Optional[int] = None,
        pid_calibration_path: Optional[str] = None,
        enable_maxigauge: bool = False,
        emulate: bool = False,
        verify_ssl: bool = False,
        timeout: float = 5.0,
        auto_pid_calibration: bool = True,
        **kwargs: Any,
    ) -> None:
        super().__init__(name=name, **kwargs)

        if folder_path:
            self.log.info(
                "The folder_path argument is ignored because this driver now "
                "talks to the controller over HTTP."
            )

        temperature_entries: List[Tuple[str, Optional[int], str]] = [
            ("50k_plate", channel_50k_plate, "50 K plate"),
            ("4k_plate", channel_4k_plate, "4 K plate"),
            ("still", channel_still, "still"),
            ("mixing_chamber", channel_mixing_chamber, "mixing chamber"),
            ("magnet", channel_magnet, "magnet"),
            ("fse", channel_fse, "FSE"),
        ]

        pressure_entries: List[Tuple[str, Optional[int], str]] = [
            ("vacuum_can", channel_vacuum_can, "vacuum can"),
            ("pumping_line", channel_pumping_line, "pumping line"),
            ("compressor_outlet", channel_compressor_outlet, "compressor outlet"),
            ("compressor_inlet", channel_compressor_inlet, "compressor inlet"),
            ("mixture_tank", channel_mixture_tank, "mixture tank"),
            ("venting_line", channel_venting_line, "venting line"),
        ]

        has_pressure_channels = any(channel is not None for _, channel, _ in pressure_entries)
        enable_maxigauge = enable_maxigauge or has_pressure_channels

        self._api = _BlueForsAPIClient(
            host=host,
            api_key=api_key,
            port=port,
            mixing_chamber_channel_id=channel_mixing_chamber,
            mixing_chamber_heater_id=mixing_chamber_heater_id,
            pid_calibration_path=pid_calibration_path,
            enable_maxigauge=enable_maxigauge,
            emulate=emulate,
            verify_ssl=verify_ssl,
            timeout=timeout,
            auto_pid_calibration=auto_pid_calibration,
            logger=self.log,
        )

        for parameter_name, channel, description in temperature_entries:
            self._register_temperature_parameter(parameter_name, channel, description)

        for parameter_name, channel, description in pressure_entries:
            self._register_pressure_parameter(parameter_name, channel, description)

        self._has_heater_controls = (
            mixing_chamber_heater_id is not None and channel_mixing_chamber is not None
        )
        if self._has_heater_controls:
            self._register_heater_parameters()

        self.connect_message()

    def get_idn(self) -> Dict[str, str]:
        """
        Return a static identification dictionary so QCoDeS does not attempt to
        send SCPI *IDN? queries, which are unsupported by the HTTP API.
        """
        return {
            "vendor": "BlueFors",
            "model": "TemperatureControllerHTTP",
            "serial": self._api._base_url,
            "firmware": "MeasureIt BlueFors HTTP driver",
        }

    def _register_temperature_parameter(
        self, parameter_name: str, channel: Optional[int], description: str
    ) -> None:
        if channel is None:
            return
        self.add_parameter(
            name=f"temperature_{parameter_name}",
            unit="K",
            get_parser=float,
            get_cmd=partial(self.get_temperature, channel),
            docstring=f"Temperature of the {description}.",
        )
        self.add_parameter(
            name=f"resistance_{parameter_name}",
            unit="Ohm",
            get_parser=float,
            get_cmd=partial(self.get_resistance, channel),
            docstring=f"Sensor resistance of the {description}.",
        )

    def _register_pressure_parameter(
        self, parameter_name: str, channel: Optional[int], description: str
    ) -> None:
        if channel is None:
            return
        self.add_parameter(
            name=f"pressure_{parameter_name}",
            unit="mBar",
            get_parser=float,
            get_cmd=partial(self.get_pressure, channel),
            docstring=f"Pressure of the {description}.",
        )

    def _register_heater_parameters(self) -> None:
        self.add_parameter(
            "mxc_heater_status",
            unit="",
            get_cmd=self._get_mxc_heater_status,
            set_cmd=self._set_mxc_heater_status,
            vals=Bool(),
            docstring="True if the mixing chamber heater output is active.",
        )
        self.add_parameter(
            "mxc_heater_pid_mode",
            unit="",
            get_cmd=self._get_mxc_heater_mode,
            set_cmd=self._set_mxc_heater_mode,
            vals=Bool(),
            docstring="PID mode toggle for the mixing chamber heater.",
        )
        self.add_parameter(
            "mxc_heater_power",
            unit="uW",
            get_cmd=self._get_mxc_heater_power,
            set_cmd=self._set_mxc_heater_power,
            vals=Numbers(min_value=0, max_value=5000),
            docstring="Mixing chamber heater power in microwatts.",
        )
        self.add_parameter(
            "mxc_heater_setpoint_mK",
            unit="mK",
            get_cmd=self._get_mxc_heater_setpoint,
            set_cmd=self._set_mxc_heater_setpoint,
            vals=Numbers(min_value=0, max_value=1000),
            docstring="Mixing chamber heater temperature setpoint in millikelvin.",
        )
        for axis in ("p", "i", "d"):
            self.add_parameter(
                f"mxc_heater_pid_{axis}",
                unit="",
                get_cmd=partial(self._get_mxc_pid_term, axis),
                set_cmd=partial(self._set_mxc_pid_term, axis),
                docstring=f"{axis.upper()} term of the mixing chamber heater PID loop.",
            )

    def _safe_api_call(self, func: Callable[[], Any], failure_value: Any) -> Any:
        try:
            return func()
        except Exception as err:  # pragma: no cover - hardware/IO dependent
            self.log.warning("BlueFors API call failed: %s", err)
            return failure_value

    def _execute_api(self, func: Callable[[], Any]) -> None:
        try:
            func()
        except Exception as err:  # pragma: no cover - hardware/IO dependent
            self.log.warning("BlueFors API command failed: %s", err)

    def get_temperature(self, channel: int) -> float:
        return float(
            self._safe_api_call(
                lambda: self._api.get_channel_temperature(channel),
                np.nan,
            )
        )

    def get_resistance(self, channel: int) -> float:
        return float(
            self._safe_api_call(
                lambda: self._api.get_channel_resistance(channel),
                np.nan,
            )
        )

    def get_pressure(self, channel: int) -> float:
        return float(
            self._safe_api_call(
                lambda: self._api.get_maxigauge_channel(channel),
                np.nan,
            )
        )

    def get_mxc_temperature(self) -> float:
        return float(self._safe_api_call(self._api.get_mxc_temperature, np.nan))

    def get_mxc_resistance(self) -> float:
        return float(self._safe_api_call(self._api.get_mxc_resistance, np.nan))

    def set_mxc_heater_setpoint(
        self, temperature_mK: float, use_pid_calibration: Optional[bool] = None
    ) -> bool:
        return bool(
            self._safe_api_call(
                lambda: self._api.set_mxc_heater_setpoint(
                    temperature_mK, use_pid_calib=use_pid_calibration
                ),
                False,
            )
        )

    def set_mxc_heater_power(self, power_uW: float) -> bool:
        return bool(
            self._safe_api_call(
                lambda: self._api.set_mxc_heater_power(power_uW), False
            )
        )

    def get_mxc_heater_pid_config(self) -> List[float]:
        return self._safe_api_call(self._api.get_mxc_heater_pid_config, [np.nan] * 3)

    def set_mxc_heater_pid_config(
        self, p: Optional[float] = None, i: Optional[float] = None, d: Optional[float] = None
    ) -> bool:
        return bool(
            self._safe_api_call(
                lambda: self._api.set_mxc_heater_pid_config(p, i, d), False
            )
        )

    def _get_mxc_heater_status(self) -> bool:
        return bool(self._safe_api_call(self._api.get_mxc_heater_status, False))

    def _set_mxc_heater_status(self, value: bool) -> None:
        self._execute_api(lambda: self._api.set_mxc_heater_status(bool(value)))

    def _get_mxc_heater_mode(self) -> bool:
        return bool(self._safe_api_call(self._api.get_mxc_heater_mode, False))

    def _set_mxc_heater_mode(self, value: bool) -> None:
        self._execute_api(lambda: self._api.set_mxc_heater_mode(bool(value)))

    def _get_mxc_heater_power(self) -> float:
        return float(self._safe_api_call(self._api.get_mxc_heater_power, np.nan))

    def _set_mxc_heater_power(self, value: float) -> None:
        self._execute_api(lambda: self._api.set_mxc_heater_power(float(value)))

    def _get_mxc_heater_setpoint(self) -> float:
        return float(self._safe_api_call(self._api.get_mxc_heater_setpoint, np.nan))

    def _set_mxc_heater_setpoint(self, value: float) -> None:
        self._execute_api(lambda: self._api.set_mxc_heater_setpoint(float(value)))

    def _get_mxc_pid_term(self, axis: str) -> float:
        return float(
            self._safe_api_call(
                lambda: self._api.get_mxc_heater_value(f"pid_{axis}"),
                np.nan,
            )
        )

    def _set_mxc_pid_term(self, axis: str, value: float) -> None:
        self._execute_api(
            lambda: self._api.set_mxc_heater_value(f"pid_{axis}", float(value))
        )



from qcodes import Instrument


class nDsweeper(Instrument):
    """Driver for combining two Keithley 2450 or NI daq channels and convert them using a linear transformation.
    Example: control top gate and bottom gate by two Keithleys and sweep carrier density or electric field which are linear combination of two Keithley output.
    parameters: [a,b,c,d]
    para1 = a*input1 + b*input2
    para2 = c*input1 + d*input2
    """

    def __init__(self, name: str, para_list: list, parameters: list, **kwargs) -> None:
        super().__init__(name, **kwargs)
        # extract device
        [self.in_para1, self.in_para2] = para_list
        [self.a, self.b, self.c, self.d] = parameters
        temp = self.a * self.d - self.b * self.c
        self.ia = self.d / temp
        self.ib = -self.b / temp
        self.ic = -self.c / temp
        self.id = self.a / temp

        self.add_parameter(
            name="para1",
            label="n",
            unit="a.u.",
            get_cmd=self._get_para1,
            get_parser=float,
            set_cmd=self._set_para1,
        )

        self.add_parameter(
            name="para2",
            label="D",
            unit="a.u.",
            get_cmd=self._get_para2,
            get_parser=float,
            set_cmd=self._set_para2,
        )

    def _get_para1(self):
        input1 = self.in_para1.get()
        input2 = self.in_para2.get()
        (val, _) = self._encoder(input1, input2)
        return val

    def _get_para2(self):
        input1 = self.in_para1.get()
        input2 = self.in_para2.get()
        (_, val) = self._encoder(input1, input2)
        return val

    def _set_para1(self, setpoint):
        # get para2
        para2 = self.para2.get()
        (input1, input2) = self._decoder(setpoint, para2)
        self.in_para1.set(input1)
        self.in_para2.set(input2)
        return None

    def _set_para2(self, setpoint):
        para1 = self.para1.get()
        (input1, input2) = self._decoder(para1, setpoint)
        self.in_para1.set(input1)
        self.in_para2.set(input2)
        return None

    """
    para1 = a*input1 + b*input2
    para2 = c*input1 + d*input2

    input1 = ia*para1 + ib*para2
    input2 = ic*para1 + id*para2
    """

    def _encoder(self, val1, val2):
        para1 = self.a * val1 + self.b * val2
        para2 = self.c * val1 + self.d * val2
        return (para1, para2)

    def _decoder(self, para1, para2):
        val1 = self.ia * para1 + self.ib * para2
        val2 = self.ic * para1 + self.id * para2
        return (val1, val2)



from qcodes import (
    VisaInstrument,
)


class LR_700(VisaInstrument):
    """QCoDeS driver for the stepped attenuator
    Weinschel is formerly known as Aeroflex/Weinschel
    """

    # all instrument constructors should accept **kwargs and pass them on to
    # super().__init__
    def __init__(self, name, address, **kwargs):
        # supplying the terminator means you don't need to remove it from every response
        super().__init__(name, address, **kwargs)

        self.add_parameter(
            "getRaw",
            # the value you set will be inserted in this command with
            # regular python string substitution. This instrument wants
            # an integer zero-padded to 2 digits. For robustness, don't
            # assume you'll get an integer input though - try to allow
            # floats (as opposed to {:0=2d})
            get_cmd=self._get_R,
            # setting any attenuation other than 0, 2, ... 60 will error.
            # vals=vals.Enum(*np.arange(0, 60.1, 2).tolist()),
            # the return value of get() is a string, but we want to
            # turn it into a (float) number
        )

        # it's a good idea to call connect_message at the end of your constructor.
        # this calls the 'IDN' parameter that the base Instrument class creates for
        # every instrument (you can override the `get_idn` method if it doesn't work
        # in the standard VISA form for your instrument) which serves two purposes:
        # 1) verifies that you are connected to the instrument
        # 2) gets the ID info so it will be included with metadata snapshots later.
        self.connect_message()

    def _get_R(self):
        self.write_raw("GET 0")
        st = self.ask_raw("")
        # while st =='\n':
        #    self.write_raw('GET 0')
        #    st = self.ask_raw('')
        # return st

        m = st.find("R")
        while True:
            if m != -1:
                break
            self.write_raw("GET 0")
            st = self.ask_raw("")
            m = st.find("R")

        # split the string and convert to float value
        sp2 = st.split(" ")
        if sp2[1] == "MOHM":
            ratio = 1e-3
        elif sp2[1] == "OHM":
            ratio = 1
        elif sp2[1] == "KOHM":
            ratio = 1000
        elif sp2[1] == "UOHM":
            ratio = 1000000

        r0 = float(sp2[0])
        r = r0 * ratio
        return r


# instantiating and using this instrument (commented out because I can't actually do it!)
#
# from qcodes.instrument_drivers.weinschel.Weinschel_8320 import Weinschel_8320
# weinschel = Weinschel_8320('w8320_1', 'TCPIP0::172.20.2.212::inst0::INSTR')
# weinschel.attenuation(40)
